__all__ = ("DcrPrecisionMetric",)

import numpy as np
import rubin_scheduler.utils as utils

import rubin_sim.maf.utils as mafUtils

from .base_metric import BaseMetric


class DcrPrecisionMetric(BaseMetric):
    """Determine how precise a DCR correction could be made

    Parameters
    ----------
    atm_err : `float`
        Minimum error in photometry centroids introduced by the atmosphere
        (arcseconds). Default 0.01.
    """

    def __init__(
        self,
        metric_name="DCRprecision",
        seeing_col="seeingFwhmGeom",
        m5_col="fiveSigmaDepth",
        ha_col="HA",
        pa_col="paraAngle",
        filter_col="filter",
        atm_err=0.01,
        sed_template="flat",
        rmag=20.0,
        **kwargs,
    ):
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.pa_col = pa_col
        self.seeing_col = seeing_col
        self.mags = {}
        self.filters = ["u", "g", "r", "i", "z", "y"]
        if sed_template == "flat":
            for f in self.filters:
                self.mags[f] = rmag
        else:
            self.mags = utils.stellarMags(sed_template, rmag=rmag)
        cols = [
            "ra_dcr_amp",
            "dec_dcr_amp",
            seeing_col,
            m5_col,
            filter_col,
            "zenithDistance",
            pa_col,
        ]
        units = "arcseconds"
        self.atm_err = atm_err
        super(DcrPrecisionMetric, self).__init__(cols, metric_name=metric_name, units=units, **kwargs)

    def run(self, data_slice, slice_point=None):
        snr = np.zeros(len(data_slice), dtype="float")
        for filt in self.filters:
            in_filt = np.where(data_slice[self.filter_col] == filt)
            snr[in_filt] = mafUtils.m52snr(self.mags[filt], data_slice[self.m5_col][in_filt])

        position_errors = np.sqrt(
            mafUtils.astrom_precision(data_slice[self.seeing_col], snr) ** 2 + self.atm_err**2
        )

        x_coord = np.tan(np.radians(data_slice["zenithDistance"])) * np.sin(
            np.radians(data_slice[self.pa_col])
        )
        x_coord2 = np.tan(np.radians(data_slice["zenithDistance"])) * np.cos(
            np.radians(data_slice[self.pa_col])
        )
        # Things should be the same for RA and dec.
        # Now I want to compute the error if I interpolate/extrapolate to +/-1.

        # function is of form, y=ax. a=y/x. da = dy/x.
        # Only strictly true if we know the unshifted position.
        # But this should be a reasonable approx.
        slope_uncerts = position_errors / x_coord
        slope_uncerts2 = position_errors / x_coord2

        total_slope_uncert = 1.0 / np.sqrt(np.sum(1.0 / slope_uncerts**2) + np.sum(1.0 / slope_uncerts2**2))

        # So, this will be the uncertainty in the RA or Dec offset at
        # x= +/- 1. A.K.A., the uncertainty in the slope
        # of the line made by tan(zd)*sin(PA) vs RA offset
        # or the line tan(zd)*cos(PA) vs Dec offset
        # Assuming we know the unshfted position of the object
        # (or there's little covariance if we are fitting for both)
        result = total_slope_uncert

        return result



__all__ = ("PhotometricParameters", "DustValues")

import os

import numpy as np
from rubin_scheduler.data import get_data_dir

from .bandpass import Bandpass
from .sed import Sed


class DustValues:
    """Calculate extinction values

    Parameters
    ----------
    R_v : `float`
        Extinction law parameter (3.1).
    bandpassDict : `dict`
        A dict with keys of filtername and values of
        rubin_sim.phot_utils.Bandpass objects.
        Default of None will load the standard ugrizy bandpasses.
    ref_ev : `float`
        The reference E(B-V) value to use. Things in MAF assume 1.

    Note
    ----
    The value that dust_values calls "ax1" is equivalent  to r_x in any filter.
    And  r_x * ebv = A_x (the extinction due to dust in any bandpass).
    DustValues.r_x is also provided as a copy of DustValues.ax1 ..
    eventually ax1 may be deprecated in favor of r_x.
    """

    def __init__(self, r_v=3.1, bandpass_dict=None, ref_ebv=1.0):
        # Calculate dust extinction values
        self.ax1 = {}
        if bandpass_dict is None:
            bandpass_dict = {}
            root_dir = os.path.join(get_data_dir(), "throughputs", "baseline")
            for f in ["u", "g", "r", "i", "z", "y"]:
                bandpass_dict[f] = Bandpass()
                bandpass_dict[f].read_throughput(os.path.join(root_dir, f"total_{f}.dat"))

        for filtername in bandpass_dict:
            wavelen_min = bandpass_dict[filtername].wavelen.min()
            wavelen_max = bandpass_dict[filtername].wavelen.max()
            testsed = Sed()
            testsed.set_flat_sed(wavelen_min=wavelen_min, wavelen_max=wavelen_max, wavelen_step=1.0)
            self.ref_ebv = ref_ebv
            # Calculate non-dust-extincted magnitude
            flatmag = testsed.calc_mag(bandpass_dict[filtername])
            # Add dust
            a, b = testsed.setup_ccm_ab()
            testsed.add_dust(a, b, ebv=self.ref_ebv, r_v=r_v)
            # Calculate difference due to dust when EBV=1.0
            # (m_dust = m_nodust - Ax, Ax > 0)
            self.ax1[filtername] = testsed.calc_mag(bandpass_dict[filtername]) - flatmag
        # Add the R_x term, to start to transition toward this name.
        self.r_x = self.ax1.copy()


def make_dict(value, bandpass_names=("u", "g", "r", "i", "z", "y", "any")):
    newdict = {}
    for f in bandpass_names:
        newdict[f] = value
    return newdict


class DefaultPhotometricParameters:
    """
    This class will just contain a bunch of dict which store
    the default PhotometricParameters for LSST Bandpasses

    Users should not access this class (which is why it is
    not included in the __all__ declaration for this file).

    It is only used to initialize PhotometricParameters for
    a bandpass name.
    """

    # Obviously, some of these parameters (effarea, gain, platescale,
    # darkcurrent, and readnoise) will not change as a function of bandpass;
    # we are just making them dicts here to be consistent with
    # everything else (and to make it possible for
    # PhotometricParameters to access them using the bandpass name
    # passed to its constructor)
    #
    # Note: all dicts contain an 'any' key which will be the default
    # value if an unknown bandpass is asked for
    #
    # 'any' values should be kept consistent with r band

    bandpass_names = ["u", "g", "r", "i", "z", "y", "any"]

    # exposure time in seconds
    exptime_sec = 15.0
    exptime = make_dict(exptime_sec)

    # number of exposures
    nexp_n = 2
    nexp = make_dict(nexp_n)

    # effective area in cm^2
    effarea_cm2 = np.pi * (6.423 / 2.0 * 100) ** 2
    effarea = make_dict(effarea_cm2)

    # electrons per ADU
    gain_adu = 2.3
    gain = make_dict(gain_adu)

    # electrons per pixel per exposure
    readnoise_e = 8.8
    readnoise = make_dict(readnoise_e)

    # electrons per pixel per second
    darkcurrent_e = 0.2
    darkcurrent = make_dict(darkcurrent_e)

    # electrons per pixel per exposure
    othernoise_e = 0.0
    othernoise = make_dict(othernoise_e)

    # arcseconds per pixel
    platescale_as = 0.2
    platescale = make_dict(platescale_as)

    # systematic squared error in magnitudes
    # see Table 14 of the SRD document
    # https://docushare.lsstcorp.org/docushare/dsweb/Get/LPM-17
    sigma_sys = {
        "u": 0.0075,
        "g": 0.005,
        "r": 0.005,
        "i": 0.005,
        "z": 0.0075,
        "y": 0.0075,
        "any": 0.005,
    }


class PhotometricParameters:
    def __init__(
        self,
        exptime=None,
        nexp=None,
        effarea=None,
        gain=None,
        readnoise=None,
        darkcurrent=None,
        othernoise=None,
        platescale=None,
        sigma_sys=None,
        bandpass=None,
    ):
        """Store photometric parameters for SNR calculations.

        Parameters
        ----------
        exptime : `float`
            Exposure time in seconds (per exposure).
            None will default to value from DefaultPhotometricParameters.
        nexp : `int`
            Number of exposures per visit.
            None will default to value from DefaultPhotometricParameters.
        effarea : `float`
            Effective area in cm^2.
            None will default to value from DefaultPhotometricParameters.
        gain : `float`
            Electrons per ADU.
            None will default to value from DefaultPhotometricParameters.
        readnoise : `float`
            Electrons per pixel per exposure.
            None will default to value from DefaultPhotometricParameters.
        darkcurrent : `float`
            Electons per pixel per second.
            None will default to value from DefaultPhotometricParameters.
        othernoise : `float`
            Electrons per pixel per exposure.
            None will default to value from DefaultPhotometricParameters.
        platescale : `float`
            Arcseconds per pixel.
            None will default to value from DefaultPhotometricParameters.
        sigma_sys : `float`
            Systematic error in magnitudes.
            None will default to value from DefaultPhotometricParameters.
        bandpass : `str`
            The name of the bandpass for these parameters.

        Examples
        --------
        If `bandpass` is set to an LSST bandpass,
        the constructor will initialize
        PhotometricParameters to LSST default values for that bandpass,
        excepting any parameters that have been set by hand.  e.g.

        >>> myPhotParams = PhotometricParameters(nexp=3, bandpass='u')

        will initialize a PhotometricParameters object to `u` band defaults,
        except with 3 exposures instead of 2. A bandpass value of None
        will use defaults from LSST `r` band where appropriate.
        """
        # readnoise, darkcurrent and othernoise are measured in electrons.
        # This is taken from the specifications document LSE-30 on Docushare
        # Section 3.4.2.3 states that the total noise per pixel shall
        # be 12.7 electrons per visit which the defaults sum to
        # (remember to multply darkcurrent by the number of seconds
        # in an exposure=15). [9 e- per 15 second exposure]

        self._exptime = None
        self._nexp = None
        self._effarea = None
        self._gain = None
        self._platescale = None
        self._sigma_sys = None
        self._readnoise = None
        self._darkcurrent = None
        self._othernoise = None

        self._bandpass = bandpass
        defaults = DefaultPhotometricParameters()

        if bandpass is None:
            bandpass_key = "any"
            # This is so we do not set the self._bandpass member variable
            # without the user's explicit consent, but we can still access
            # default values from the PhotometricParameterDefaults
        else:
            bandpass_key = bandpass

        if bandpass_key in defaults.bandpass_names:
            self._exptime = defaults.exptime[bandpass_key]
            self._nexp = defaults.nexp[bandpass_key]
            self._effarea = defaults.effarea[bandpass_key]
            self._gain = defaults.gain[bandpass_key]
            self._platescale = defaults.platescale[bandpass_key]
            self._sigma_sys = defaults.sigma_sys[bandpass_key]
            self._readnoise = defaults.readnoise[bandpass_key]
            self._darkcurrent = defaults.darkcurrent[bandpass_key]
            self._othernoise = defaults.othernoise[bandpass_key]

        if exptime is not None:
            self._exptime = exptime

        if nexp is not None:
            self._nexp = nexp

        if effarea is not None:
            self._effarea = effarea

        if gain is not None:
            self._gain = gain

        if platescale is not None:
            self._platescale = platescale

        if sigma_sys is not None:
            self._sigma_sys = sigma_sys

        if readnoise is not None:
            self._readnoise = readnoise

        if darkcurrent is not None:
            self._darkcurrent = darkcurrent

        if othernoise is not None:
            self._othernoise = othernoise

        failure_message = ""
        failure_ct = 0

        if self._exptime is None:
            failure_message += "did not set exptime\n"
            failure_ct += 1

        if self._nexp is None:
            failure_message += "did not set nexp\n"
            failure_ct += 1

        if self._effarea is None:
            failure_message += "did not set effarea\n"
            failure_ct += 1

        if self._gain is None:
            failure_message += "did not set gain\n"
            failure_ct += 1

        if self._platescale is None:
            failure_message += "did not set platescale\n"
            failure_ct += 1

        if self._sigma_sys is None:
            failure_message += "did not set sigma_sys\n"
            failure_ct += 1

        if self._readnoise is None:
            failure_message += "did not set readnoise\n"
            failure_ct += 1

        if self._darkcurrent is None:
            failure_message += "did not set darkcurrent\n"
            failure_ct += 1

        if self._othernoise is None:
            failure_message += "did not set othernoise\n"
            failure_ct += 1

        if failure_ct > 0:
            raise RuntimeError("In PhotometricParameters:\n%s" % failure_message)

    @property
    def bandpass(self):
        """
        The name of the bandpass associated with these parameters.
        Can be None.
        """
        return self._bandpass

    @bandpass.setter
    def bandpass(self, value):
        raise RuntimeError(
            "You should not be setting bandpass on the fly; "
            + "Just instantiate a new case of PhotometricParameters"
        )

    @property
    def exptime(self):
        """
        exposure time in seconds
        """
        return self._exptime

    @exptime.setter
    def exptime(self, value):
        raise RuntimeError(
            "You should not be setting exptime on the fly; "
            + "Just instantiate a new case of PhotometricParameters"
        )

    @property
    def nexp(self):
        """
        number of exposures
        """
        return self._nexp

    @nexp.setter
    def nexp(self, value):
        raise RuntimeError(
            "You should not be setting nexp on the fly; "
            + "Just instantiate a new case of PhotometricParameters"
        )

    @property
    def effarea(self):
        """
        effective area in cm^2
        """
        return self._effarea

    @effarea.setter
    def effarea(self, value):
        raise RuntimeError(
            "You should not be setting effarea on the fly; "
            + "Just instantiate a new case of PhotometricParameters"
        )

    @property
    def gain(self):
        """
        electrons per ADU
        """
        return self._gain

    @gain.setter
    def gain(self, value):
        raise RuntimeError(
            "You should not be setting gain on the fly; "
            + "Just instantiate a new case of PhotometricParameters"
        )

    @property
    def platescale(self):
        """
        arcseconds per pixel
        """
        return self._platescale

    @platescale.setter
    def platescale(self, value):
        raise RuntimeError(
            "You should not be setting platescale on the fly; "
            + "Just instantiate a new case of PhotometricParameters"
        )

    @property
    def readnoise(self):
        """
        electrons per pixel per exposure
        """
        return self._readnoise

    @readnoise.setter
    def readnoise(self, value):
        raise RuntimeError(
            "You should not be setting readnoise on the fly; "
            + "Just instantiate a new case of PhotometricParameters"
        )

    @property
    def darkcurrent(self):
        """
        electrons per pixel per second
        """
        return self._darkcurrent

    @darkcurrent.setter
    def darkcurrent(self, value):
        raise RuntimeError(
            "You should not be setting darkcurrent on the fly; "
            + "Just instantiate a new case of PhotometricParameters"
        )

    @property
    def othernoise(self):
        """
        electrons per pixel per exposure
        """
        return self._othernoise

    @othernoise.setter
    def othernoise(self, value):
        raise RuntimeError(
            "You should not be setting othernoise on the fly; "
            + "Just instantiate a new case of PhotometricParameters"
        )

    @property
    def sigma_sys(self):
        """
        systematic error in magnitudes
        """
        return self._sigma_sys

    @sigma_sys.setter
    def sigma_sys(self, value):
        raise RuntimeError(
            "You should not be setting sigma_sys on the fly; "
            + "Just instantiate a new case of PhotometricParameters"
        )



__all__ = ("get_imsim_flux_norm",)

import numpy as np

from .bandpass import Bandpass


def get_imsim_flux_norm(sed, magmatch):
    """
    Calculate the flux normalization of an SED in the imsim bandpass.

    Parameters
    -----------
    sed is the SED to be normalized

    magmatch is the desired magnitude in the imsim bandpass

    Returns
    --------
    The factor by which the flux of sed needs to be multiplied to achieve
    the desired magnitude.
    """

    # This method works based on the assumption that the imsim bandpass
    # is a delta function.  If that ever ceases to be true, the unit test
    # testSedUtils.py, which checks that the results of this method are
    # identical to calling Sed.calcFluxNorm and passing in the imsim bandpass,
    # will fail and we will know to modify this method.

    if not hasattr(get_imsim_flux_norm, "imsim_wavelen"):
        bp = Bandpass()
        bp.imsim_bandpass()
        non_zero_dex = np.where(bp.sb > 0.0)[0][0]
        get_imsim_flux_norm.imsim_wavelen = bp.wavelen[non_zero_dex]

    if sed.fnu is None:
        sed.flambda_tofnu()

    if (
        get_imsim_flux_norm.imsim_wavelen < sed.wavelen.min()
        or get_imsim_flux_norm.imsim_wavelen > sed.wavelen.max()
    ):
        raise RuntimeError(
            "Cannot normalize sed "
            "at wavelength of %e nm\n" % get_imsim_flux_norm.imsim_wavelen
            + "The SED does not cover that wavelength\n"
            + "(Covers %e < lambda %e)" % (sed.wavelen.min(), sed.wavelen.max())
        )

    mag = -2.5 * np.log10(np.interp(get_imsim_flux_norm.imsim_wavelen, sed.wavelen, sed.fnu)) - sed.zp
    dmag = magmatch - mag
    return np.power(10, (-0.4 * dmag))



"""Second generation hourglass plotting classes.
"""

# pylint: disable=too-many-arguments
# pylint: disable=super-init-not-called
# pylint: disable=too-many-instance-attributes
# pylint: disable=too-few-public-methods
__all__ = (
    "GeneralHourglassPlot",
    "CategoricalHourglassPlotMixin",
    "RangeHourglassCategoricalPlot",
    "MonthHourglassCategoricalPlot",
    "MonthHourglassUsePlot",
    "TimeUseHourglassPlotMixin",
    "MonthHourglassPlot",
    "YearHourglassCategoricalPlot",
    "YearHourglassPlot",
    "YearHourglassUsePlot",
)

import calendar
import copy
import logging
from collections import OrderedDict

import astropy.coordinates
import astropy.time
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from astropy import units as u
from rubin_scheduler.utils.ddf_locations import ddf_locations
from rubin_scheduler.utils.riseset import riseset_times

from .plot_handler import BasePlotter


class GeneralHourglassPlot(BasePlotter):
    """Make an hourglass plot

    Parameters
    ----------
    tz : `str`
        The timezone to use
    site : `str`
        The site name (sent to `astropy.coordinates.EarthLocation.of_site`)
    solar_time : `bool`
        Use solar time as the x axis (instead of civil time)?
    marked_ra : `dict`
        A dictionary of RA values (in deg) for which to label transit lines.

    A general feature of the hourglass plot is that you can pass more
    data (such as the metric values, calculated for the entire survey)
    to the plotter than are actually included in the plot. The plotter
    will pull out the select subset of data, defined by kwargs.

    """

    def __init__(
        self,
        tz="Chile/Continental",
        site="Cerro Pachon",
        solar_time=True,
        marked_ra=None,
    ):
        self.plot_type = "GeneralizedHourglass"
        self.object_plotter = False  # pylint: disable=invalid-name
        self.default_plot_dict = {
            "title": None,
            "xlabel": "Hours after midnight",
            "ylabel": "MJD",
            "x_min": -6,
            "x_max": 6,
            "figsize": None,
            "cmap": plt.get_cmap("viridis"),
            "colorbar": True,
            "legend": False,
            "legend_ncols": 2,
            "legend_bbox_to_anchor": (0.9, 0.0),
            "legend_loc": "lower right",
        }
        self.plot_dict = copy.copy(self.default_plot_dict)  # pylint: disable=invalid-name
        self.tz = tz  # pylint: disable=invalid-name
        if isinstance(site, str):
            self.site = astropy.coordinates.EarthLocation.of_site(site)
        else:
            self.site = site

        self.solar_time = solar_time
        self.color_map = {}
        self.marked_ra = {} if marked_ra is None else marked_ra

    def _add_axis_labels(self, ax):  # pylint: disable=invalid-name, no-self-use
        """Add title/x/y labels to one set of axes.

        Parameters
        ----------
        ax : `matplotlib.axes.Axes`
           The axis to which to add labels
        """
        # Isolate code to add axis titles and labels because we need
        # to be able to call it independently of the actually plotting,
        # so that when we plot grids of axes on the same figure
        # we don't have overlaps and repeats.
        ax.set_title(self.plot_dict["title"])
        ax.set_xlabel(self.plot_dict["xlabel"])
        ax.set_ylabel(self.plot_dict["ylabel"])

    def _plot(self, fig, intervals):
        """Create axes for the data and add the data itself.

        This method does not add the legend or color bar. Use the
        __call__ method to combine all needed steps: handling creation
        of the intervals DataFrame, calling this method, and then
        adding any necessary legends or colorbars.

        Parameters
        ----------
        fig : `matplotlib.figure.Figure`
           figure on which to plot
        intervals : `pandas.DataFrame`
           data to be plotted, with "mjd", "duration", and "value" columns.
           "duration" should be in units of hours.

        Returns
        -------
        color_mappable : `matplotlib.cm.ScalarMappable`
           map of scalar data to colors
        axes : numpy.array
           a numpy array of matplotlib axes.

        """

        # This is similar to the __call__ method
        # in other plotters, but isolates the making of the plot itself
        # from other things __call__ needs to do, such as reformat the
        # data and add figure-wide elements like the legend and color
        # bar, so that these tasks can be replaced independently in
        # subclases without needing to repeat the other elements of
        # call, which might not change.

        # Subclasses should override this method to change
        # how the plotting itself is done, as distinct from
        # restructuring the data or modifying other elements of the
        # figure such as the color bar and legend.

        ax = fig.add_axes([0, 0, 1, 1])  # pylint: disable=invalid-name
        color_mappable = self._plot_one_axis(intervals, ax)
        self._add_axis_labels(ax, self.plot_dict)
        return [color_mappable, np.array([ax])]

    def _plot_one_axis(self, intervals, ax, epoch_mjd=None):  # pylint: disable=invalid-name
        """Actually add the data to one axis of an hourglass plot

        includes adding the twilight/transit lines, does not include any
        supporting infrastructure or plot elements: it does not create
        the figure or axis, or add labels, color bars, or other
        elements of the plot that do not represent the data itself.

        Parameters
        ----------
        intervals : `pandas.DataFrame`
           data to be plotted, with "mjd", "duration", and "value" columns.
           "duration" should be in units of hours.
        ax : `matplotlib.axes.Axes`
           the axes on which to plot data
        epoch_mjd : `float`
           the MJD used to set y limits for the plot
        """

        # Add content relevant to an individual axis.
        # It is separated from _plot so that it can be called
        # in different subplots, while _plot drives the the content
        # of all subplots.

        # Add the 'block' data
        if len(intervals) > 0:
            if self.solar_time:
                (
                    night_mjds,
                    hours_after_midnight,
                ) = _compute_hours_after_solar_midnight(intervals["mjd"].values, self.site)
            else:
                (
                    night_mjds,
                    hours_after_midnight,
                ) = _compute_hours_after_midnight(intervals["mjd"].values)

            color_mappable = self._plot_value_bars(
                intervals["value"],
                night_mjds,
                hours_after_midnight,
                intervals["duration"],
                epoch_mjd,
                ax,
            )
        else:
            color_mappable = None

        try:
            ax.set_ylim(self.plot_dict["y_max"], self.plot_dict["y_min"])
        except KeyError:
            ax.set_ylim(ax.get_ylim()[-1] + 1, -0.5)
        ax.set_xlim(self.plot_dict["x_min"], self.plot_dict["x_max"])

        start_mjd = ax.get_ylim()[1] + epoch_mjd
        end_mjd = ax.get_ylim()[0] + epoch_mjd
        # Add the twilight boundaries and moon transit lines
        _astron_hourglass(start_mjd + 0.5, end_mjd + 1, ax, self.site, self.solar_time)
        # Add the transits
        self._plot_transits(start_mjd + 0.5, end_mjd + 1, ax)

        return color_mappable

    def _plot_value_bars(  # pylint: disable=invalid-name
        self,
        values,
        night_mjds,
        hours_after_midnight,
        duration,
        epoch_mjd,
        ax,
    ):
        """Add the color bar blocks for the metric values.

        Parameters
        ----------
        values : `numpy.array`
           values to assign to the value bars
        night_mjds : `numpy.array`
           array if integer MJDs of the nights
           for each bar.
        hours_after_midnight : `numpy.array`
           array of times (hours after midnight) for the
           start times of each bar.
        duration : `numpy.array`
           array of durations (in hours) for each bar
        epoch_mjd : `int`
           the reference MJD of the vertical axis
           of the plot, for example the start of the
           month if the vertical axis is to be labeled
           with the day of month.
        ax : `matplotlib.axes.Axes`
           the axes on which to plot the bars

        Returns
        -------
        color_mappable : `matplotlib.cm.ScalarMappable`
           relates values to colors for the bars
           (see the matplotlib documentation)
        """

        # Drives the plotting of the elements that represent the metric itself,
        # changing as a function of time.

        # Find the colors to be used.
        colors, color_mappable = self._map_colors(values)

        # Add the bars.
        ax.bar(
            hours_after_midnight,
            height=0.8,
            width=duration,
            bottom=night_mjds - epoch_mjd - 0.4,
            color=colors,
            linewidth=0,
            align="edge",
        )

        return color_mappable

    def _map_colors(self, values):  # pylint: disable=invalid-name, no-self-use
        """Get colors corresponding to values

        Paramaters
        ----------
        values : `numpy.array`
           Values for which to get colors

        Returns
        -------
        colors : `numpy.array`
           colors corresponding to the provided values
        color_mappable : `matplotlib.cm.ScalarMappable`
           relates values to colors for the bars
           (see the matplotlib documentation)
        """

        # Manages the mapping of metric values to colors in the
        # horizonatal lines in the hourglass plot.
        cmap = self.plot_dict["cmap"]
        if values.max() == values.min():
            colors = cmap(0.5)
        else:
            try:
                color_limits = self.plot_dict["color_limits"]
            except KeyError:
                color_limits = (values.min(), values.max())
            colors = cmap((values - color_limits[0]) / (color_limits[1] - color_limits[0]))

        norm = mpl.colors.Normalize(vmin=color_limits[0], vmax=color_limits[1])
        color_mappable = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        color_mappable.set_array([])

        return colors, color_mappable

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        """Restructure the metric data to use, and build the figure.

        Parameters
        ----------
        metric_value : `numpy.ma.MaskedArray`
            The metric values from the bundle.
        slicer : `rubin_sim.maf.slicers.TwoDSlicer`
            The slicer.
        user_plot_dict: `dict`
            Dictionary of plot parameters set by user
            (overrides default values).
        fig : `matplotlib.figure.Figure`
            Matplotlib figure number to use. Default = None, starts new figure.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
           Figure with the plot.
        """
        # Highest level driver for the plotter.
        # Prepares data structures and figure-wide elements

        # Building the figure entails plotting it the relevant
        # class._plot function and adding plot decorations with
        # the relevant class._add_colorbar and class._add_figure_legend
        # methods.

        intervals = pd.DataFrame(
            {
                "mjd": slicer.slice_points["mjd"],
                "duration": (slicer.slice_points["duration"] * u.s).to_value(
                    u.hour  # pylint: disable=no-member
                ),
                "value": metric_value,
            }
        )
        # Set up a copy of the plot dict to contain entries relevant
        # for this plot
        self.plot_dict = copy.copy(self.default_plot_dict)  # pylint: disable=invalid-name
        self.plot_dict.update(user_plot_dict)

        # Generate the figure
        if fig is None:
            fig = plt.figure(figsize=self.plot_dict["figsize"])

        # Add the plots
        color_mappable, axes = self._plot(fig, intervals)

        # add the colorbar, if requested
        if self.plot_dict["colorbar"] and color_mappable is not None:
            self._add_colorbar(fig, color_mappable, axes)

        # add the legend, if requested
        if self.plot_dict["legend"] and len(self.color_map) > 0:
            fig = self._add_figure_legend(fig, axes)

        return fig

    def _add_figure_legend(self, fig, axes):
        """Creates and adds the figure legend.

        Parameters
        ----------
        fig : `matplotlib.figure.Figure`
           figure on which to plot
        axes : `matplotlib.axes.Axes`
           axes from which to extract
           automatic labels.

        Returns
        -------
        fig : `matplotlib.figure.Figure`
           The matplotlib figure
        """
        # Creates and adds the figure legend.  This method follows two
        # stages: first, it adds entries for each element in the color
        # map (defined in self.color_map).  Then, it adds any
        # additional elements of the plats that were given matplotlib
        # labels, but are not already in the legend by way of the
        # color map.  The user may limit the enties in the color map
        # by way of the legend_elements key in the plot_dict, which if
        # preset limits the elements of the key to only those
        # requested.

        # Build a dictionary and build lable and handle lists
        # later, so that we can ensure that the order of the
        # labels in the keys matches the order specified by the
        # user, if present.
        handle_dict = OrderedDict()

        # Label everything we have assigned a color
        for label in self.color_map:
            if "legend_elements" in self.plot_dict:
                if label not in self.plot_dict["legend_elements"]:
                    continue

            handle_dict[label] = mpl.patches.Patch(facecolor=self.color_map[label], label=label)

        # If we missed anything we have assigned a label while
        # plotting, automatically pick it up, but avoid
        # duplication
        auto_handles, auto_labels = axes.T.flatten()[-1].get_legend_handles_labels()
        for auto_handle, auto_label in zip(auto_handles, auto_labels):
            handle_dict[auto_label] = auto_handle

        # If the user specified which elements they wanted labels,
        # include only them (if present).
        # Otherwise, unclude all defined labels
        if "legend_elements" in self.plot_dict:
            desired_labels = self.plot_dict["legend_elements"]
        else:
            desired_labels = handle_dict.keys()

        labels = [lbl for lbl in desired_labels if lbl in handle_dict]
        handles = [handle_dict[lbl] for lbl in desired_labels if lbl in handle_dict]

        fig.legend(
            handles,
            labels,
            loc=self.plot_dict["legend_loc"],
            ncol=self.plot_dict["legend_ncols"],
            bbox_to_anchor=self.plot_dict["legend_bbox_to_anchor"],
        )

        return fig

    def _add_colorbar(self, fig, color_mappable, axes):  # pylint: disable=invalid-name, no-self-use
        """Add a colorbar.

        Parameters
        ----------
        fig : `matplotlib.figure.Figure`
           figure on which to plot
        color_mappable : `matplotlib.cm.ScalarMappable`
           map of scalar data to colors
        axes : `matplotlib.axes.Axes`
           The axes for which to add the colorbar
        """

        colorbar_kwargs = {"fraction": 0.05}
        if "colorbar_ticks" in self.plot_dict:
            colorbar_kwargs["ticks"] = self.plot_dict["colorbar_ticks"]
        fig.colorbar(color_mappable, ax=axes, **colorbar_kwargs)

    def _plot_dates(  # pylint: disable=invalid-name
        self,
        intervals,
        start_tstamp,
        end_tstamp,
        ax,
        epoch_tstamp=None,
    ):
        """Plot data on one axis for a range of dates.

        Parameters
        ----------
        intervals : `pandas.DataFrame`
           data to be plotted, with "mjd", "duration", and "value" columns.
           "duration" should be in units of hours.
        start_tstamp : `pandas.Timestamp`
           The first time to include in the axes
        end_tstamp : `pandas.Timestamp`
           The last time to include in the axes
        ax : `matplotlib.axes.Axes`
           the axes on which to plot data
        epoch_tstamp : `float`
           the time used to set y limits for the plot


        Returns
        -------
        color_mappable : `matplotlib.cm.ScalarMappable`
           map of data to colors
        """
        if self.plot_dict is None:
            self.plot_dict = {}

        start_mjd = start_tstamp.to_julian_date() - 2400000.5
        end_mjd = end_tstamp.to_julian_date() - 2400000.5
        epoch_mjd = 0 if epoch_tstamp is None else epoch_tstamp.to_julian_date() - 2400000.5

        these_intervals = intervals.query(f"{start_mjd}<mjd<{end_mjd}")
        color_mappable = self._plot_one_axis(these_intervals, ax, epoch_mjd)
        return color_mappable

    def _plot_month(self, intervals, month, year, ax):
        """Plots data as function of date in the month.

        intervals : `pandas.DataFrame`
           data to be plotted, with "mjd", "duration", and "value" columns.
           "duration" should be in units of hours.
        month : `int`
           Month number whose dates to plot
        year : `int`
           Year of the month to plot
        ax : `matplotlib.axes.Axes`
           the axes on which to plot data

        Returns
        -------
        color_mappable : `matplotlib.cm.ScalarMappable`
           map of data to colors
        """

        if self.plot_dict is None:
            self.plot_dict = {}

        start_tstamp = pd.Timestamp(year=year, month=month, day=1, hour=12, tz=self.tz)
        end_tstamp = start_tstamp + pd.DateOffset(months=1)
        epoch_tstamp = pd.Timestamp(year=year, month=month, day=1, hour=0, tz="UTC") - pd.DateOffset(days=1)
        color_mappable = self._plot_dates(
            intervals,
            start_tstamp,
            end_tstamp,
            ax,
            epoch_tstamp=epoch_tstamp,
        )
        ax.set_yticks(np.arange(1, 32, 7))
        return color_mappable

    def _plot_transits(self, start_mjd, end_mjd, ax):  # pylint: disable=invalid-name
        """Plot transit lines on a set of axes.

        Parameters
        ----------
        start_mjd : `int`
           Start time for the axes
        end_mjd: `int`
           End time for the exes
        ax: `matplotlib.axes.Axes`
           the axes on which to plot transit lines
        """
        # Add transit lines for each of the RA values
        # it self.marked_ra.
        mjds = np.arange(start_mjd, end_mjd + 1, 1)

        for field in self.marked_ra:
            ra = self.marked_ra[field]  # pylint: disable=invalid-name
            try:
                color = self.color_map[field]
            except KeyError:
                continue

            transit_mjds = _compute_coord_transit_mjds(mjds, ra, self.site)

            if self.solar_time:
                (
                    transit_nights,
                    transit_hams,
                ) = _compute_hours_after_solar_midnight(transit_mjds, self.site)
            else:
                transit_nights, transit_hams = _compute_hours_after_midnight(transit_mjds)

            ax.plot(
                transit_hams,
                transit_nights - start_mjd + 1,
                color=color,
                label=field + "-transit",
            )


#
# Plot appearance subclasses
#

# The following subclasses change the arrangement of dates on the
# hourglass plot.


class RangeHourglassPlot(GeneralHourglassPlot):
    """Make an hourglass plot for a range of dates

    Parameters
    ----------
    start_date : `str`
        The start date (in a format parseable by `pandas.Timestamp`)
    end_date : `str`
        The end date (in a format parseable by `pandas.Timestamp`)

    Keyward arguments are passed to `GeneralHourglassPlot`
    """

    def __init__(self, start_date, end_date, **kwargs):
        super().__init__(**kwargs)
        self.plot_type = "RangeHourglass"
        self.start_date = pd.Timestamp(start_date)
        self.end_date = pd.Timestamp(end_date)

    def _plot(self, fig, intervals):
        """Make the hourglass plot for a range of dates.

        This method does not add the legend or color bar. Use the
        __call__ method to handle creation of the intervals DataFrame,
        call this method, and then add any necessary legends or
        colorbars.

        Parameters
        ----------
        fig : `matplotlib.figure.Figure`
           figure on which to plot
        intervals : `pandas.DataFrame`
           data to be plotted, with "mjd", "duration", and "value" columns.
           "duration" should be in units of hours.

        """
        ax = fig.add_axes([0, 0, 1, 1])  # pylint: disable=invalid-name
        color_mappable = self._plot_dates(intervals, self.start_date, self.end_date, ax)
        self._add_axis_labels(ax, self.plot_dict)
        return color_mappable, np.array([ax])


class MonthHourglassPlot(GeneralHourglassPlot):
    """Make an hourglass plot for a month

    Parameters
    ----------
    month : `int`
        The month number (1-12).
    year : `int`
        The year.

    Keyward arguments are passed to `GeneralHourglassPlot`

    Note that this pulls the chosen month's data out of the metric
    values calculated for the entire survey.  Unsubclassed, this class
    expects to obtain float/int data.  To use it with other data, its
    subclass that includes the appropriate mixin
    (MonthHourglassCategoricalPlot or MonthHourglassUsePlot) or a
    custom subclass.

    """

    def __init__(self, month, year, **kwargs):
        super().__init__(**kwargs)
        self.plot_type = "MonthHourglass"
        self.default_plot_dict["ylabel"] = "day of month"
        self.default_plot_dict["y_max"] = calendar.monthrange(year, month)[1] + 0.5
        self.default_plot_dict["y_min"] = 0.5
        self.month = month
        self.year = year

    def _plot(self, fig, intervals):
        """Make the hourglass plot for a month.

        This method does not add the legend or color bar. Use the
        __call__ method to handle creation of the intervals DataFrame,
        call this method, and then add any necessary legends or
        colorbars.

        Parameters
        ----------
        fig : `matplotlib.figure.Figure`
           figure on which to plot
        intervals : `pandas.DataFrame`
           data to be plotted, with "mjd", "duration", and "value" columns.
           "duration" should be in units of hours.

        """

        ax = fig.add_axes([0, 0, 1, 1])  # pylint: disable=invalid-name
        color_mappable = self._plot_month(intervals, self.month, self.year, ax)
        self._add_axis_labels(ax)
        return color_mappable, np.array([ax])


class YearHourglassPlot(GeneralHourglassPlot):
    """Make an array of monthly hourglass plots for a year.

    Parameters
    ----------
    year : `int`
        The year.

    Keyward arguments are passed to `GeneralHourglassPlot`

    Note that this plot pulls the chosen year's data out of the metric
    values calculated for the entire survey.  Like the
    MonthHourglassPlot class, this class expects to obtain float/int
    data.  To use it with other data, its subclass that includes the
    appropriate mixin (YearHourglassCategoricalPlot or
    YearHourglassUsePlot) or a custom subclass.

    """

    def __init__(self, year, **kwargs):
        super().__init__(**kwargs)
        self.year = year
        self.plot_type = "YearHourglass"
        self.default_plot_dict["ylabel"] = "day of month"
        self.default_plot_dict["y_max"] = 31.5
        self.default_plot_dict["y_min"] = 0.5

    def _plot(self, fig, intervals):
        """Make a series of month plots, arranged for the entire year.

        This method does not add the legend or color bar. Use the
        __call__ method to handle creation of the intervals DataFrame,
        call this method, and then add any necessary legends or
        colorbars.

        Parameters
        ----------
        fig : `matplotlib.figure.Figure`
           figure on which to plot
        intervals : `pandas.DataFrame`
           data to be plotted, with "mjd", "duration", and "value" columns.
           "duration" should be in units of hours.

        """

        axes = fig.subplots(
            3,
            4,
            sharex=True,
            sharey=True,
            gridspec_kw={"wspace": 0.025, "hspace": 0},
        )

        for month, ax in zip(np.arange(1, 13), axes.T.flatten()):  # pylint: disable=invalid-name
            logging.info("Working on %s, %d", calendar.month_name[month], self.year)
            self.plot_dict["y_max"] = calendar.monthrange(self.year, month)[1] + 0.5
            color_mappable = self._plot_month(intervals, month, self.year, ax)
            ax.set_title("")
            ax.set_title(calendar.month_abbr[month], y=1, x=0.005, pad=-15, loc="left")

        self._add_figure_labels(fig, axes)
        return color_mappable, axes.ravel()

    def _add_figure_labels(self, fig, axes):  # pylint: disable=invalid-name, no-self-use
        """Generate the figure labels.

        Parameters
        ----------
        fig : `matplotlib.figure.Figure`
           figure to which to add lables
        ax : `matplotlib.axes.Axes`
           The axis to which to add labels
        """
        axes[1, 0].set_ylabel(self.plot_dict["ylabel"])
        fig.suptitle(self.plot_dict["title"], y=0.9)
        fig.text(0.5, 0.05, self.plot_dict["xlabel"], ha="center")


#
# Mixin classes to change handling of metric data
#

# Mixin classes avoid the need to duplicate code for every appearance
# type subclass. Rather that need to duplicate code between
# RangeHourglassPlot, MonthHourglassPlot, and YearHourglassPlot to
# enable them all to support hourglass data, a single "mixin" class
# allows the needed modifications to be defined once, and then "mixed
# in" to each of these classes (see RangeHourglassCategoricalPlot,
# MonthHourglassCategoricalPlot, and YearHourglassCategoricalPlot
# below).


class CategoricalHourglassPlotMixin:
    """A mix-in to allow the HourglassPlot to accept categorical data,
    rather than float/int.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.object_plotter = True  # pylint: disable=invalid-name
        self.plot_type = (  # pylint: disable=invalid-name
            "Categorical" + self.plot_type
        )  # pylint: disable=invalid-name
        self.default_plot_dict["cmap"] = plt.get_cmap("tab10")
        self.default_plot_dict["assigned_colors"] = OrderedDict()
        self.default_plot_dict["legend"] = True
        self.default_plot_dict["colorbar"] = False
        if "marked_ra" not in kwargs:
            self.marked_ra = {}

    def _map_colors(self, values):
        cmap = self.plot_dict["cmap"]
        assigned_colors = self.plot_dict["assigned_colors"]
        self.color_map = {}
        if len(self.color_map.keys()) == 0:
            self.color_map = _assign_category_colors(
                np.unique(values),
                cmap,
                assigned_colors=assigned_colors,
            )

        colors = np.array(np.vectorize(self.color_map.get)(values)).T.tolist()
        color_mappable = None

        return colors, color_mappable


class RangeHourglassCategoricalPlot(CategoricalHourglassPlotMixin, RangeHourglassPlot):
    """Plot categorical data for a range of dates."""


class MonthHourglassCategoricalPlot(CategoricalHourglassPlotMixin, MonthHourglassPlot):
    """Plot categorical data for a Month."""


class YearHourglassCategoricalPlot(CategoricalHourglassPlotMixin, YearHourglassPlot):
    """Plot categorical data for a year."""


class TimeUseHourglassPlotMixin(CategoricalHourglassPlotMixin):
    """A mix-in to allow the HourglassPlot to accept categorical 'use' data
    rather than float/int."""

    def __init__(self, *args, **kwargs):
        """Customize hourglass plotter for time use"""
        super().__init__(*args, **kwargs)
        self.plot_type = "Use" + self.plot_type  # pylint: disable=invalid-name
        self.default_plot_dict["cmap"] = plt.get_cmap("tab10")
        self.default_plot_dict["colorbar"] = False
        self.default_plot_dict["assigned_colors"] = OrderedDict(
            (
                ("wide with only IR", 3),
                ("wide with u, g, or r", 0),
                ("greedy", 1),
            )
        )

        self.default_plot_dict["legend"] = True
        if "marked_ra" not in kwargs:
            self.marked_ra = {f: c[0] for f, c in ddf_locations().items()}


class MonthHourglassUsePlot(TimeUseHourglassPlotMixin, MonthHourglassPlot):
    """Plot categorical 'use' data for one month."""


class YearHourglassUsePlot(TimeUseHourglassPlotMixin, YearHourglassPlot):
    """Plot categorical 'use' data for one year."""


# internal functions & classes


def _assign_category_colors(uses, cmap, use_colors=None, assigned_colors=None):
    """Set a dictionary of nice colors for the use blocks.
    Options allow specifing pre-defined elements for some categories."""

    use_colors = OrderedDict() if use_colors is None else use_colors
    assigned_colors = OrderedDict() if assigned_colors is None else assigned_colors

    available_idxs = list(range(cmap.N))

    used_colors = set(use_colors.values())
    for cmap_idx in range(cmap.N):
        # Skip gray because it is confusing against background
        this_cmap = cmap(cmap_idx)
        if this_cmap[0] == this_cmap[1] and this_cmap[0] == this_cmap[2]:
            available_idxs.remove(cmap_idx)

        # If a color has been used already, don't reuse it for something else
        if cmap(cmap_idx) in used_colors:
            available_idxs.remove(cmap_idx)

    for use, cmap_idx in assigned_colors.items():
        if cmap_idx in available_idxs:
            available_idxs.remove(cmap_idx)
            use_colors[use] = cmap(cmap_idx)
        else:
            assert use_colors[use] == cmap(cmap_idx)

    for use in uses:
        if use not in use_colors:
            use_idx = available_idxs[0]
            use_colors[use] = cmap(use_idx)
            available_idxs.remove(use_idx)

    return use_colors


def _compute_hours_after_midnight(mjd, tz="Chile/Continental"):  # pylint: disable=invalid-name
    """Calculate the civil 'hours' value for the hourglass plot.

    Parameters
    ----------
    mjd : `numpy.array`
       array of Modified Julian Dates
    tz : `str`
       The timezone (to be passed to
       `pandas.DatetimeIndex.tz_convret`)

    Returns
    -------
    night_mjds : `numpy.array`
       array of interger nights for the dates of the provided mjds.
    hours_after_midnight : `numpy.array`
       array of hours after local civil midnight in the timezone
       provided, for each input mjd value.
    """

    utc_datetimes = pd.to_datetime(mjd + 2400000.5, unit="D", origin="julian").tz_localize("UTC")
    local_times = pd.DatetimeIndex(utc_datetimes).tz_convert(tz)
    night_mjds = np.floor(local_times.to_julian_date() - 2400001).astype(int).values
    hours_after_midnight = (local_times.to_julian_date().values - 2400001.5 - night_mjds) * 24
    return night_mjds, hours_after_midnight


def _compute_hours_after_solar_midnight(mjd, site, tz="Chile/Continental"):  # pylint: disable=invalid-name
    """Calculate the solar 'hours' value for the hourglass plot.

    Parameters
    ----------
    mjd : `numpy.array`
       array of Modified Julian Dates
    tz : `str`
       The timezone (to be passed to
       `pandas.DatetimeIndex.tz_convret`)

    Returns
    -------
    night_mjds : `numpy.array`
       array of interger nights for the dates of the provided mjds.
    hours_after_midnight : `numpy.array`
       array of hours after mean local solar midnight (antitransit),
       for each input mjd value.
    """

    times = astropy.time.Time(mjd, format="mjd", location=site)
    mean_solar_jd = times.ut1.mjd + site.lon.deg / 360
    mean_solar_time = astropy.coordinates.Angle(
        mean_solar_jd * 360, unit=u.deg  # pylint: disable=no-member
    ).wrap_at(
        180 * u.deg  # pylint: disable=no-member
    )
    hours_after_midnight = (
        mean_solar_time.to_value(u.deg) * 24 / 360.0  # pylint: disable=no-member  # pylint: disable=no-member
    )  # pylint: disable=no-member
    utc_datetimes = pd.to_datetime(mjd + 2400000.5, unit="D", origin="julian").tz_localize("UTC")
    local_times = pd.DatetimeIndex(utc_datetimes).tz_convert(tz)
    night_mjds = np.floor(local_times.to_julian_date() - 2400001).astype(int).values
    return night_mjds, hours_after_midnight


def _compute_coord_transit_mjds(mjds, ra, site):  # pylint: disable=invalid-name
    """Compute the coordinate transit times.

    Parameters
    ----------
    mjds : `numpy.array`
       array of Modified Julian Dates for the nights
       for which to calculate transit
    ra : `float`
       R.A. for which to calculate transit
       in degrees
    site : `astropy.coordinates.EarthLocation`
       location of the telescope site

    Return
    ------
    mjds : `numpy.array`
       array of MJDs of transit times.
    """
    # Used for transit lines for the hourglass plot.

    # Calculate the local sidereal time at mjds - in degrees (for RA
    # comparison)
    lsts = astropy.time.Time(mjds, format="mjd", location=site).sidereal_time("apparent").deg
    # pylint: disable=invalid-name, no-member
    ha = astropy.coordinates.Angle(lsts - ra, unit=u.deg).wrap_at(180 * u.deg)
    mjds = mjds - ha.to_value(u.deg) / 360
    return mjds


def _compute_moon_transit_mjds(mjds, site):
    """Compute the transit times for the moon,

    Parameters
    ----------
    mjds : `numpy.array`
       array of Modified Julian Dates for the nights
       for which to calculate transit
    site : `astropy.coordinates.EarthLocation`
       location of the telescope site

    Return
    ------
    mjds : `numpy.array`
       array of MJDs of transit times.
    """
    # for transit lines for the hourglass plot.

    for _ in np.arange(3):
        times = astropy.time.Time(mjds, format="mjd", location=site)
        moon_coords = astropy.coordinates.get_body("moon", times)
        # Calculate the local sidereal time at mjds
        lsts = astropy.time.Time(mjds, format="mjd", location=site).sidereal_time("apparent")
        # pylint: disable=invalid-name, no-member
        ha = (lsts - moon_coords.ra).wrap_at(180 * u.deg)
        mjds = mjds - ha.to_value(u.deg) / 360
    return mjds


def _astron_hourglass(
    start_mjd, end_mjd, ax, site, solar_time
):  # pylint: disable=invalid-name, too-many-locals, too-many-branches
    """Add the moon transit, night and twilight start/end lines.

    Parameters
    ----------
    start_mjd : `int`
       Start time for the axes
    end_mjd: `int`
       End time for the exes
    ax: `matplotlib.axes.Axes`
       the axes on which to plot transit lines
    site : `astropy.coordinates.EarthLocation`
       location of the telescope site
    solar_time : `bool`
       True to use mean local solar midnight
       False to use civil midnight

    Returns
    -------
    ax: `matplotlib.axes.Axes`
       the axes on transit lines were plotted
    """

    mjds = np.arange(start_mjd, end_mjd + 1, 1)
    start_mjd = mjds[0]

    # Moon transit

    cal_night_mjds = mjds - site.lon.deg / 360

    moon_transit_mjds = _compute_moon_transit_mjds(cal_night_mjds, site)

    if solar_time:
        (
            moon_transit_nights,
            moon_transit_hams,
        ) = _compute_hours_after_solar_midnight(moon_transit_mjds, site)
    else:
        moon_transit_nights, moon_transit_hams = _compute_hours_after_midnight(moon_transit_mjds)
    # Loop to avoid wrapping
    moon_lines = np.cumsum(np.diff(moon_transit_hams, prepend=moon_transit_hams[0]) < 0)
    moon_label = "moon"
    for moon_line in np.unique(moon_lines):
        these_hams = moon_transit_hams[moon_lines == moon_line]
        these_nights = moon_transit_nights[moon_lines == moon_line]
        ax.plot(
            these_hams,
            these_nights - start_mjd + 1,
            color="yellow",
            linewidth=8,
            alpha=0.5,
            label=moon_label,
        )
        moon_label = None

    # Moon rise and set
    for direction in ("up", "down"):
        moon_event_mjds = riseset_times(cal_night_mjds, direction, alt=0, body="moon")
        if solar_time:
            (
                moon_event_nights,
                moon_event_hams,
            ) = _compute_hours_after_solar_midnight(moon_event_mjds, site)
        else:
            moon_event_nights, moon_event_hams = _compute_hours_after_midnight(moon_event_mjds)
        # Loop to avoid wrapping
        moon_lines = np.cumsum(np.diff(moon_event_hams, prepend=moon_event_hams[0]) < 0)
        for moon_line in np.unique(moon_lines):
            these_hams = moon_event_hams[moon_lines == moon_line]
            these_nights = moon_event_nights[moon_lines == moon_line]
            ax.plot(
                these_hams,
                these_nights - start_mjd + 1,
                color="yellow",
                linestyle="dotted",
            )

    # Twilight
    guess_offset = {"up": 0.2, "down": -0.2}
    twilight_shade = {0: "0.7", -6: "0.6", -12: "0.4", -18: "black"}

    for alt in twilight_shade:
        pm_event_mjds = riseset_times(cal_night_mjds + guess_offset["down"], "down", alt=alt, body="sun")
        am_event_mjds = riseset_times(cal_night_mjds + guess_offset["up"], "up", alt=alt, body="sun")
        if solar_time:
            (
                pm_event_nights,
                pm_event_hams,
            ) = _compute_hours_after_solar_midnight(pm_event_mjds, site)
            (
                am_event_nights,
                am_event_hams,
            ) = _compute_hours_after_solar_midnight(am_event_mjds, site)
        else:
            (
                pm_event_nights,
                pm_event_hams,
            ) = _compute_hours_after_midnight(pm_event_mjds)
            (
                am_event_nights,
                am_event_hams,
            ) = _compute_hours_after_midnight(am_event_mjds)
        assert np.array_equal(am_event_nights, pm_event_nights)
        ax.fill_betweenx(
            am_event_nights - start_mjd + 1,
            pm_event_hams,
            am_event_hams,
            color=twilight_shade[alt],
            zorder=-10,
        )

    return ax



import numpy as np

import rubin_sim.maf.metrics as metrics
from rubin_sim.maf.maf_contrib.lss_obs_strategy.galaxy_counts_metric_extended import (
    GalaxyCountsMetricExtended as GalaxyCountsMetric,
)


class DepthLimitedNumGalMetric(metrics.BaseMetric):
    """This metric calculates the number of galaxies while accounting for the
    extragalactic footprint.

    Parameters
    ----------
    m5_col : `str`, optional
        Name of column for depth in the data. Default: 'fiveSigmaDepth'
    filter_col : `str`, optional
        Name of column for filter in the data. Default: 'filter'
    maps : `list` [`str`], optional
        List of map names. Default: ['DustMap']
    nside : `int`, optional
        HEALpix resolution parameter. Default: 256.
        This should match slicer nside.
    filter_band : `str`, optional
        Filter to use to calculate galaxy counts.
        Any one of 'u', 'g', 'r', 'i', 'z', 'y'. Default: 'i'
    redshiftBin: `str`, optional
        options include '0.<z<0.15', '0.15<z<0.37', '0.37<z<0.66, '0.66<z<1.0',
        '1.0<z<1.5', '1.5<z<2.0', '2.0<z<2.5', '2.5<z<3.0','3.0<z<3.5',
        '3.5<z<4.0', 'all' for no redshift restriction
        (so consider 0.<z<4.0)
        Default: 'all'
    nfilters_needed : `int`, optional
        Number of filters in which to require coverage. Default: 6
    lim_mag_i_ptsrc : `float`, optional
        Point-source limiting mag for the i-band coadded dust-corrected depth.
        Default: 26.0
    lim_ebv : `float`, optional
        Limiting EBV value. Default: 0.2

    Returns
    -------
    ngal : `float`
        Number of galaxies in healpix if the slice_point passes the
        extragalactic cuts; otherwise self.badval
    """

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        nside=128,
        filter_band="i",
        redshift_bin="all",
        nfilters_needed=6,
        lim_mag_i_ptsrc=26.0,
        lim_ebv=0.2,
        metric_name="DepthLimitedNumGalaxiesMetric",
        units="Galaxy Counts",
        **kwargs,
    ):
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.filter_band = filter_band
        # set up the extended source limiting mag
        # galaxies are x2 as seeing: seeing is generally 0.7arcsec
        # and a typical galaxies is 1arcsec
        # => for extended source limiting mag of x, we'd need x + 0.7
        # as the point-source limiting mag;
        # 0.7 comes from $\sqrt{1/2}$;
        # basically have x2 difference in magnitudes between
        # point source and extended source.
        lim_mag_i_extsrc = lim_mag_i_ptsrc - 0.7
        # set up the metric for galaxy counts
        self.galmetric = GalaxyCountsMetric(
            m5_col=self.m5_col,
            nside=nside,
            upper_mag_limit=lim_mag_i_extsrc,
            include_dust_extinction=True,
            filter_band=self.filter_band,
            redshift_bin=redshift_bin,
            cfht_ls_counts=False,
            normalized_mock_catalog_counts=True,
        )
        # set up the metric for extragalactic footprint
        self.eg_metric = metrics.ExgalM5WithCuts(
            m5_col=self.m5_col,
            filter_col=self.filter_col,
            lsst_filter=self.filter_band,
            n_filters=nfilters_needed,
            extinction_cut=lim_ebv,
            depth_cut=lim_mag_i_ptsrc,
        )
        maps = self.eg_metric.maps + self.galmetric.maps
        maps = set(maps)
        super().__init__(
            col=[self.m5_col, self.filter_col], maps=maps, metric_name=metric_name, units=units, **kwargs
        )

    def run(self, data_slice, slice_point=None):
        # see if this slice_point is in the extragalactic footprint
        pass_egcuts = self.eg_metric.run(data_slice, slice_point=slice_point)

        if pass_egcuts == self.badval:  # failed dust/depth cuts
            return self.badval

        # Otherwise, find the galaxy counts
        in_filt = np.where(data_slice[self.filter_col] == self.filter_band)[0]
        return self.galmetric.run(data_slice[in_filt], slice_point=slice_point)



import os
import shutil
import tempfile
import unittest

from rubin_scheduler.data import get_data_dir
from rubin_scheduler.utils.code_utilities import sims_clean_up

import rubin_sim.maf as maf

TEST_DB = "example_v3.4_0yrs.db"


class Test3x2(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        sims_clean_up()

    def setUp(self):
        self.out_dir = tempfile.mkdtemp(prefix="TMB")

    @unittest.skipUnless(
        os.path.isdir(os.path.join(get_data_dir(), "maps")),
        "Skipping 3x3 metric test because no dust maps.",
    )
    def test_3x2(self):
        # Only testing that the metric successfully runs, not checking that
        # the output values are valid.
        bundle_list = []
        nside = 64
        colmap = maf.batches.col_map_dict("fbs")
        nfilters_needed = 6
        lim_ebv = 0.2
        ptsrc_lim_mag_i_band = 25.9
        m = maf.metrics.ExgalM5WithCuts(
            m5_col=colmap["fiveSigmaDepth"],
            filter_col=colmap["filter"],
            lsst_filter="i",
            n_filters=nfilters_needed,
            extinction_cut=lim_ebv,
            depth_cut=ptsrc_lim_mag_i_band,
        )
        s = maf.slicers.HealpixSlicer(nside=nside, use_cache=False)
        sql = 'note not like "DD%" and night < 365'
        threeby_two_summary_simple = maf.metrics.StaticProbesFoMEmulatorMetricSimple(
            nside=nside, metric_name="3x2ptFoM_simple"
        )
        threeby_two_summary = maf.maf_contrib.StaticProbesFoMEmulatorMetric(
            nside=nside, metric_name="3x2ptFoM"
        )
        bundle_list.append(
            maf.metric_bundles.MetricBundle(
                m,
                s,
                sql,
                summary_metrics=[threeby_two_summary, threeby_two_summary_simple],
            )
        )

        database = os.path.join(get_data_dir(), "tests", TEST_DB)
        results_db = maf.db.ResultsDb(out_dir=self.out_dir)
        bd = maf.metric_bundles.make_bundles_dict_from_list(bundle_list)
        bg = maf.metric_bundles.MetricBundleGroup(bd, database, out_dir=self.out_dir, results_db=results_db)
        bg.run_all()

    def tearDown(self):
        if os.path.isdir(self.out_dir):
            shutil.rmtree(self.out_dir)


if __name__ == "__main__":
    unittest.main()



import os
import unittest

from rubin_scheduler.data import get_data_dir

from rubin_sim.phot_utils import Bandpass


class ReadBandPassTest(unittest.TestCase):
    """
    Tests for reading in bandpasses
    """

    def test_read(self):
        """
        Check that we can read things stored in the throughputs directory.
        """
        throughputs_dir = os.path.join(get_data_dir(), "throughputs")

        # select files to try and read
        files = [
            "2MASS/2MASS_Ks.dat",
            "WISE/WISE_w1.dat",
            "johnson/johnson_U.dat",
            "sdss/sdss_r.dat",
        ]
        for filename in files:
            bp = Bandpass()
            bp.read_throughput(os.path.join(throughputs_dir, filename))



# Add similar lines (from .filename import *) when you add new metrics,
# stackers or slicers.
from .calculate_lsst_field_visibility_astropy import *
from .depth_limited_num_gal_metric import *
from .filter_pair_t_gaps_metric import *
from .grb_transient_metric import *
from .gw170817_det_metric import *
from .kne_metrics import *
from .lss_metrics import *
from .lss_obs_strategy import *
from .lv_dwarfs import *
from .microlensing_metric import *
from .periodic_metric import *
from .periodic_star_metric import *
from .periodic_star_modulation_metric import *
from .presto_color_kne_pop_metric import *
from .selfcal_uniformity_metric import *
from .star_count_mass_metric import *
from .star_count_metric import *
from .static_probes_fom_summary_metric import *
from .tdes_pop_metric import *
from .triplet_metric import *
from .var_depth_metric import *
from .xrb_metrics import *
from .young_stellar_objects_metric import *



__all__ = ("NightPointingPlotter",)

import warnings

import matplotlib.pyplot as plt
import numpy as np

from .plot_handler import BasePlotter


class NightPointingPlotter(BasePlotter):
    def __init__(self, mjd_col="observationStartMJD", alt_col="alt", az_col="az"):
        # Just call it Hourglass so it gets treated the same way
        self.plot_type = "Hourglass"
        self.mjd_col = mjd_col
        self.alt_col = alt_col
        self.az_col = az_col
        self.object_plotter = True
        self.default_plot_dict = {
            "title": None,
            "xlabel": "MJD",
            "ylabels": ["Alt", "Az"],
            "figsize": None,
        }
        self.filter2color = {
            "u": "purple",
            "g": "blue",
            "r": "green",
            "i": "cyan",
            "z": "orange",
            "y": "red",
        }

    def __call__(self, metric_value, slicer, user_plot_dict, fig=None):
        if fig is not None:
            warnings.warn("Always expect to create a new plot in " "NightPointingPlotter.")
        fig, (ax1, ax2) = plt.subplots(2, sharex=True)

        mv = metric_value[0]

        u_filters = np.unique(mv["data_slice"]["filter"])
        for filt in u_filters:
            good = np.where(mv["data_slice"]["filter"] == filt)
            ax1.plot(
                mv["data_slice"][self.mjd_col][good],
                mv["data_slice"][self.alt_col][good],
                "o",
                color=self.filter2color[filt],
                markersize=5,
                alpha=0.5,
            )
            ax2.plot(
                mv["data_slice"][self.mjd_col][good],
                mv["data_slice"][self.az_col][good],
                "o",
                color=self.filter2color[filt],
                markersize=5,
                alpha=0.5,
            )

        good = np.where(np.degrees(mv["moon_alts"]) > -10.0)
        ax1.plot(
            mv["mjds"][good],
            np.degrees(mv["moon_alts"][good]),
            "ko",
            markersize=3,
            alpha=0.1,
        )
        ax2.plot(
            mv["mjds"][good],
            np.degrees(mv["moon_azs"][good]),
            "ko",
            markersize=3,
            alpha=0.1,
        )
        ax2.set_xlabel("MJD")
        ax1.set_ylabel("Altitude (deg)")
        ax2.set_ylabel("Azimuth (deg)")

        good = np.where(np.degrees(mv["sun_alts"]) > -20.0)
        ax1.plot(mv["mjds"][good], np.degrees(mv["sun_alts"][good]), "yo", markersize=3)
        ax2.plot(mv["mjds"][good], np.degrees(mv["sun_azs"][good]), "yo", markersize=3)

        ax1.set_ylim([-20.0, 90.0])
        ax2.set_ylim([0.0, 360.0])

        for i, key in enumerate(["u", "g", "r", "i", "z", "y"]):
            ax1.text(
                1.05,
                0.9 - i * 0.07,
                key,
                color=self.filter2color[key],
                transform=ax1.transAxes,
            )

        return fig



__all__ = ("get_data_dir", "get_baseline", "data_dict", "rs_download_testing", "rs_download_data")

import argparse
import glob
import os

from rubin_scheduler.data import DEFAULT_DATA_URL, download_rubin_data
from rubin_scheduler.data import get_data_dir as gdd


def get_data_dir() -> str:
    """Wraps rubin_scheduler.data.get_data_dir().
    Provided here for backwards compatibility.

    Returns
    -------
    $RUBIN_SIM_DATA_DIR : `str`
        Directory containing the necessary data for rubin_sim_data.
    """
    return gdd()


def get_baseline() -> str:
    """Get the path to the baseline cadence simulation sqlite file.

    Returns
    -------
    file : `str`
        Path to the baseline cadence simulation sqlite file.
    """
    dd = get_data_dir()
    path = os.path.join(dd, "sim_baseline")
    file = glob.glob(path + "/*10yrs.db")[0]
    return file


def data_dict() -> dict:
    """
    Dictionary containing expected version information for rubin_sim_data
    data sets, for this version of rubin_sim.

    Returns
    -------
    file_dict : `dict`
        Data bucket filenames dictionary with keys:
        ``"name"``
        Data bucket name (`str`).
        ``"version"``
        Versioned file name (`str`).
    """
    # Note for developers:
    # to create tar files and follow any sym links, run: e.g.
    #  ``tar -chvzf maf_may_2021.tgz maf``
    file_dict = {
        "maf": "maf_2024_06_13.tgz",
        "maps": "maps_2022_2_28.tgz",
        "movingObjects": "movingObjects_oct_2021.tgz",
        "orbits": "orbits_2022_3_1.tgz",
        "orbits_precompute": "orbits_precompute_2023_05_23.tgz",
        "sim_baseline": "sim_baseline_2024_10_08.tgz",
        "skybrightness": "skybrightness_2023_09_11.tgz",
        "throughputs": "throughputs_2023_09_22.tgz",
        "tests": "tests_2024_04_23.tgz",
    }
    return file_dict


def rs_download_testing() -> None:
    """Convenience function for github actions, to download only a subset
    of data.

    Contains maf,maps,movingObjects,skybrightness,throughputs and tests.
    For users who don't care about moving objects, this is likely sufficient.
    """
    all_files = data_dict()
    keys = ["maf", "maps", "movingObjects", "skybrightness", "throughputs", "tests"]
    subset = {}
    for k in keys:
        subset[k] = all_files[k]

    download_rubin_data(
        subset,
        dirs=None,
        print_versions_only=False,
        update=True,
        force=False,
        url_base=DEFAULT_DATA_URL,
        tdqm_disable=True,
    )


def rs_download_data() -> None:
    """Utility to download necessary data for rubin_sim.

    Wrapper around rubin_scheduler.scheduler_download_data,
    but downloading the data files specified by rubin_sim.
    """

    files = data_dict()
    parser = argparse.ArgumentParser(description="Download data files for rubin_sim package")
    parser.add_argument(
        "--versions",
        dest="versions",
        default=False,
        action="store_true",
        help="Report expected versions, then quit",
    )
    parser.add_argument(
        "-d",
        "--dirs",
        type=str,
        default=None,
        help="Comma-separated list of directories to download",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        default=False,
        action="store_true",
        help="Force re-download of data directory(ies)",
    )
    parser.add_argument(
        "--url_base",
        type=str,
        default=DEFAULT_DATA_URL,
        help="Root URL of download location",
    )
    parser.add_argument(
        "--tdqm_disable",
        dest="tdqm_disable",
        default=False,
        action="store_true",
        help="Turn off tdqm progress bar",
    )
    parser.add_argument(
        "--update",
        dest="update",
        default=False,
        action="store_true",
        help="Update versions of data on disk to match current",
    )

    args = parser.parse_args()

    download_rubin_data(
        files,
        dirs=args.dirs,
        print_versions_only=args.versions,
        update=args.update,
        force=args.force,
        url_base=args.url_base,
        tdqm_disable=args.tdqm_disable,
    )



__all__ = ("PeriodicStarModulationMetric",)

import random
import warnings

import numpy as np
from scipy.optimize import curve_fit

from rubin_sim.maf.metrics.base_metric import BaseMetric
from rubin_sim.maf.utils import m52snr

from .periodic_star_metric import PeriodicStar

""" This metric is based on the PeriodicStar metric
    It was modified in a way to reproduce attempts to identify
    phase modulation (Blazhko effect) in RR Lyrae stars.
    We are not implementing a period/ phase modulation in the light curve,
    but rather use short baselines (e.g.: 20 days) of observations to test
    how well we can recover the period, phase and amplitude.
    We do this as such an attempt is also useful for other purposes,
    i.e. if we want to test whether we can just recover period, phase
    and amplitude from short baselines at all, without necessarily having
    in mind to look for period/ phase modulations.
    Like in the PeriodicStar metric, the light curve of an RR Lyrae star,
    or a periodic star in general, is approximated as a simple sin wave.
    Other solutions might make use of light curve templates
    to generate light curves.
    Two other modifications we introduced are:
    In contrast to the PeriodicStar metric, we allow for a random phase
    offset to mimic observation starting at random phase.
    Also, we vary the periods and amplitudes within +/- 10 % to allow
    for a more realistic sample of variable stars.

    This metric is based on the cadence note:
    N. Hernitschek, K. Stassun, LSST Cadence Note:
    "Cadence impacts on reliable classification of standard-candle
    variable stars (2021)"
     https://docushare.lsst.org/docushare/dsweb/Get/Document-37673
"""


class PeriodicStarModulationMetric(BaseMetric):
    """Evaluate how well a periodic source can be fit on a short baseline,
    using a Monte Carlo simulation.

    At each slice_point, run a Monte Carlo simulation to see how well a
    periodic source can be fit.
    Assumes a simple sin-wave light-curve, and generates Gaussain noise
    based in the 5-sigma limiting depth of each observation.
    Light curves are evaluated piecewise to test how well we can recover
    the period, phase and amplitude from shorter baselines.
    We allow for a random phase offset to mimic observation starting
    at random phase. Also, we vary the periods and amplitudes
    within +/- 10 % to allow for a more realistic sample of variable stars.

    Parameters
    ----------
    period : `float`, opt
        days (default 10)
    amplitude : `float`, opt
        mags (default 0.5)
    phase : `float`, opt
        days (default 2.)
    random_phase : `bool`, opt
        a random phase is assigned (default False)
    time_interval : `float`, opt
        days (default 50);
        the interval over which we want to evaluate the light curve
    n_monte : `int`, opt
        number of noise realizations to make in the Monte Carlo (default 1000)
    period_tol : `float`, opt
        fractional tolerance on the period to demand
        for a star to be considered well-fit (default 0.05)
    amp_tol : `float`, opt
        fractional tolerance on the amplitude to demand (default 0.10)
    means : `list` of `float`, opt
        mean magnitudes for ugrizy (default all 20)
    mag_tol : `float`, opt
        Mean magnitude tolerance (mags) (default 0.1)
    n_bands : `int`, opt
        Number of bands that must be within mag_tol (default 3)
    seed : `int`, opt
        random number seed (default 42)
    """

    def __init__(
        self,
        metric_name="PeriodicStarModulationMetric",
        mjd_col="observationStartMJD",
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        period=10.0,
        amplitude=0.5,
        phase=2.0,
        random_phase=False,
        time_interval=50,
        n_monte=1000,
        period_tol=0.05,
        amp_tol=0.10,
        means=[20.0, 20.0, 20.0, 20.0, 20.0, 20.0],
        mag_tol=0.10,
        n_bands=3,
        seed=42,
        **kwargs,
    ):
        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        super(PeriodicStarModulationMetric, self).__init__(
            col=[self.mjd_col, self.m5_col, self.filter_col],
            units="Fraction Detected",
            metric_name=metric_name,
            **kwargs,
        )
        self.period = period
        self.amplitude = amplitude
        self.time_interval = time_interval
        if random_phase:
            self.phase = np.nan
        else:
            self.phase = phase
        self.n_monte = n_monte
        self.period_tol = period_tol
        self.amp_tol = amp_tol
        self.means = np.array(means)
        self.mag_tol = mag_tol
        self.n_bands = n_bands
        np.random.seed(seed)
        self.filter2index = {"u": 3, "g": 4, "r": 5, "i": 6, "z": 7, "y": 8}

    def run(self, data_slice, slice_point=None):
        # Bail if we don't have enough points
        # (need to fit mean magnitudes in each of the available bands
        # - self.means and for a period, amplitude, and phase)
        if data_slice.size < self.means.size + 3:
            return self.badval

        # Generate input for true light curve

        lightcurvelength = data_slice.size

        t = np.empty(lightcurvelength, dtype=list(zip(["time", "filter"], [float, "|U1"])))
        t["time"] = data_slice[self.mjd_col] - data_slice[self.mjd_col].min()
        t["filter"] = data_slice[self.filter_col]
        m5 = data_slice[self.m5_col]

        lightcurvelength_days = self.time_interval

        # evaluate light curves piecewise in subruns
        subruns = int(np.max(t["time"]) / lightcurvelength_days)

        # print('number of subruns: ', subruns)
        frac_recovered_list = []

        for subrun_idx in range(0, subruns):
            good = (t["time"] >= (lightcurvelength_days * (subrun_idx))) & (
                t["time"] <= (lightcurvelength_days * (subrun_idx + 1))
            )
            t_subrun = t[good]
            m5_subrun = m5[good]
            if t_subrun["time"].size > 0:
                # If we are adding a distance modulus to the magnitudes
                if "distMod" in list(slice_point.keys()):
                    mags = self.means + slice_point["distMod"]
                else:
                    mags = self.means
                # slightly different periods and amplitudes (+/- 10 %)
                # to mimic true stars. random phase offsets to mimic
                # observation starting at random phase
                true_period = random.uniform(0.9, 1.1) * self.period
                true_amplitude = random.uniform(0.9, 1.1) * self.amplitude
                if np.isnan(self.phase):
                    # a random phase (in days) should be assigned
                    true_phase = random.uniform(0, 1) * self.period
                else:
                    true_phase = self.phase

                true_params = np.append(np.array([true_period, true_phase, true_amplitude]), mags)
                true_obj = PeriodicStar(t_subrun["filter"])
                true_lc = true_obj(t_subrun["time"], *true_params)

                # Array to hold the fit results
                fits = np.zeros((self.n_monte, true_params.size), dtype=float)
                for i in np.arange(self.n_monte):
                    snr = m52snr(true_lc, m5_subrun)
                    dmag = 2.5 * np.log10(1.0 + 1.0 / snr)
                    noise = np.random.randn(true_lc.size) * dmag
                    # Suppress warnings about failing on covariance
                    fit_obj = PeriodicStar(t_subrun["filter"])
                    # check if we have enough points
                    if np.size(true_params) >= np.size(fit_obj):
                        parm_vals = true_params * 0 + np.inf
                    else:
                        with warnings.catch_warnings():
                            warnings.simplefilter("ignore")
                            # If it fails to converge,
                            # save values that should fail later
                            try:
                                parm_vals, pcov = curve_fit(
                                    fit_obj,
                                    t_subrun["time"],
                                    true_lc + noise,
                                    p0=true_params,
                                    sigma=dmag,
                                )
                            except RuntimeError:
                                parm_vals = true_params * 0 + np.inf
                    fits[i, :] = parm_vals

                # Throw out any magnitude fits if there are no
                # observations in that filter
                ufilters = np.unique(data_slice[self.filter_col])
                if ufilters.size < 9:
                    for key in list(self.filter2index.keys()):
                        if key not in ufilters:
                            fits[:, self.filter2index[key]] = -np.inf

                # Find the fraction of fits that meet the "well-fit" criteria
                period_frac_err = np.abs((fits[:, 0] - true_params[0]) / true_params[0])
                amp_frac_err = np.abs((fits[:, 2] - true_params[2]) / true_params[2])
                mag_err = np.abs(fits[:, 3:] - true_params[3:])
                n_bands = np.zeros(mag_err.shape, dtype=int)
                n_bands[np.where(mag_err <= self.mag_tol)] = 1
                n_bands = np.sum(n_bands, axis=1)
                n_recovered = np.size(
                    np.where(
                        (period_frac_err <= self.period_tol)
                        & (amp_frac_err <= self.amp_tol)
                        & (n_bands >= self.n_bands)
                    )[0]
                )
                frac_recovered = float(n_recovered) / self.n_monte
                frac_recovered_list.append(frac_recovered)

        frac_recovered = np.sum(frac_recovered_list) / (len(frac_recovered_list))
        return frac_recovered



__all__ = ("StackerRegistry", "BaseStacker")

import inspect
import warnings

import numpy as np


class StackerRegistry(type):
    """
    Meta class for Stackers, to build a registry of stacker classes.
    """

    def __init__(cls, name, bases, dict):
        super(StackerRegistry, cls).__init__(name, bases, dict)
        if not hasattr(cls, "registry"):
            cls.registry = {}
        if not hasattr(cls, "source_dict"):
            cls.source_dict = {}
        modname = inspect.getmodule(cls).__name__
        if modname.startswith("rubin_sim.maf.stackers"):
            modname = ""
        else:
            if len(modname.split(".")) > 1:
                modname = ".".join(modname.split(".")[:-1]) + "."
            else:
                modname = modname + "."
        stackername = modname + name
        if stackername in cls.registry:
            raise Exception(
                "Redefining stacker %s! (there are >1 stackers with the same name)" % (stackername)
            )
        if stackername != "BaseStacker":
            cls.registry[stackername] = cls
        cols_added = cls.cols_added
        for col in cols_added:
            cls.source_dict[col] = cls

    def get_class(cls, stackername):
        return cls.registry[stackername]

    def help(cls, doc=False):
        for stackername in sorted(cls.registry):
            if not doc:
                print(stackername)
            if doc:
                print("---- ", stackername, " ----")
                print(cls.registry[stackername].__doc__)
                stacker = cls.registry[stackername]()
                print(" Columns added to SimData: ", ",".join(stacker.cols_added))
                print(" Default columns required: ", ",".join(stacker.cols_req))


class BaseStacker(metaclass=StackerRegistry):
    """Base MAF Stacker.

    Stackers add columns generated at run-time to the simdata array.
    """

    # List of the names of the columns generated by the Stacker.
    cols_added = []

    def __init__(self):
        """
        Instantiate the stacker.
        This method should be overriden by the user.
        This serves to define the API required by MAF.
        """
        # Add the list of new columns generated by the stacker
        # as class attributes (colsAdded - above).

        # List of the names of the columns required from the database
        # (to generate the Stacker columns).
        self.cols_req = []
        # Optional: specify the new column types.
        self.cols_added_dtypes = None
        # Optional: provide a list of units for the columns
        # defined in colsAdded.
        self.units = [None]

    def __hash__(self):
        return None

    def __eq__(self, other_stacker):
        """
        Evaluate if two stackers are equivalent.

        This method is required to determine if metric_bundles may be
        evaluated at the same time, on the same data.
        """
        # If the class names are different, they are not 'the same'.
        if self.__class__.__name__ != other_stacker.__class__.__name__:
            return False
        # Otherwise, this is the same stacker class,
        # but may be instantiated differently.
        # We have to delve a little further,
        # and compare the kwargs & attributes for each stacker.
        state_now = dir(self)
        for key in state_now:
            if not key.startswith("_") and key != "registry" and key != "run" and key != "next":
                if not hasattr(other_stacker, key):
                    return False
                # If the attribute is from numpy, assume it's an array
                # and test it
                if type(getattr(self, key)).__module__ == np.__name__:
                    if not np.array_equal(getattr(self, key), getattr(other_stacker, key)):
                        return False
                else:
                    if getattr(self, key) != getattr(other_stacker, key):
                        return False
        return True

    def __ne__(self, other_stacker):
        """
        Evaluate if two stackers are not equal.
        """
        if self == other_stacker:
            return False
        else:
            return True

    def _add_stacker_cols(self, sim_data):
        """
        Add the new Stacker columns to the sim_data array.
        If columns already present in sim_data,
        just allows 'run' method to overwrite.
        Returns sim_data array with these columns added
        (so 'run' method can set their values).
        """
        if not hasattr(self, "cols_added_dtypes") or self.cols_added_dtypes is None:
            self.cols_added_dtypes = [float for col in self.cols_added]
        # Create description of new recarray.
        newdtype = sim_data.dtype.descr
        cols_present = [False] * len(self.cols_added)
        for i, (col, dtype) in enumerate(zip(self.cols_added, self.cols_added_dtypes)):
            if col in sim_data.dtype.names:
                if sim_data[col][0] is not None:
                    cols_present[i] = True
                    warnings.warn(
                        "Warning - column %s already present in sim_data, may be overwritten "
                        "(depending on stacker)." % (col)
                    )
            else:
                newdtype += [(col, dtype)]
        new_data = np.empty(sim_data.shape, dtype=newdtype)
        # Add references to old data.
        for col in sim_data.dtype.names:
            new_data[col] = sim_data[col]
        # Were all columns present and populated with something not None?
        # If so, then consider 'all there'.
        if sum(cols_present) == len(self.cols_added):
            cols_present = True
        else:
            cols_present = False
        return new_data, cols_present

    def run(self, sim_data, override=False):
        """
        Run the stacker, adding new columns.

        Parameters
        ----------
        sim_data : `np.ndarray`, (N, M)
            The data to be used to evaluate metrics.
        override : `bool`, optional
            If True, recalculate new (stacker) columns even if present.
            If False, calculates stacker columns only if they are not present.


        Returns
        -------
        sim_data : `np.ndarray`, (N, MM)
            The input data, plus additional stacker columns.
        """
        # Add new columns
        if len(sim_data) == 0:
            return sim_data
        sim_data, cols_present = self._add_stacker_cols(sim_data)
        # If override is set, it means go ahead and recalculate stacker values.
        if override:
            cols_present = False
        # Run the method to calculate/add new data.
        try:
            return self._run(sim_data, cols_present)
        except TypeError:
            warnings.warn(
                "Please update the stacker %s so that the _run method matches the current API. "
                "This will give you the option to skip re-running stackers if the columns are "
                "already present." % (self.__class__.__name__)
            )
            return self._run(sim_data)

    def _run(self, sim_data, cols_present=False):
        """Do the work to add the new columns. This method should be overriden
        in subclasses.

        Parameters
        ----------
        sim_data: `np.ndarray`, (N, M)
            The observation data, provided by the MAF framework.
        cols_present: `bool`, optional
            Flag to indicate what to do with the columns already present in
            the data.  This will also be provided by the MAF framework --
            but your _run method can use the value. If it is 'True' and you
            do trust the existing value, the _run method can simply
            return sim_data without additional calculations.

        Returns
        -------
        sim_data : `np.ndarray`, (N, MM)
            The simdata, with the columns added or updated
            (or simply already present).
        """
        # By moving the calculation of these columns to a separate method,
        # we add the possibility of using stackers with pandas dataframes.
        # The _addStackerCols method won't work with dataframes, but the
        #  _run methods are quite likely to (depending on their details),
        #  as they are just populating columns.
        raise NotImplementedError(
            "Not Implemented: " "the child stackers should implement their own _run methods"
        )



__all__ = (
    "MetricRow",
    "DisplayRow",
    "PlotRow",
    "SummaryStatRow",
    "VersionRow",
    "ResultsDb",
)

import datetime
import os
import time
import warnings
from sqlite3 import OperationalError

import numpy as np
from sqlalchemy import Column, Float, ForeignKey, Integer, String, create_engine
from sqlalchemy.engine import url
from sqlalchemy.exc import DatabaseError
from sqlalchemy.orm import backref, declarative_base, relationship, sessionmaker
from sqlalchemy.sql import text

import rubin_sim.version as rsVersion

Base = declarative_base()


class MetricRow(Base):
    """
    Define contents and format of metric list table.

    (Table to list all metrics, their info_label, and their output data files).
    """

    __tablename__ = "metrics"
    # Define columns in metric list table.
    metric_id = Column(Integer, primary_key=True)
    metric_name = Column(String)
    slicer_name = Column(String)
    run_name = Column(String)
    sql_constraint = Column(String)
    metric_info_label = Column(String)
    metric_datafile = Column(String)

    def __repr__(self):
        return (
            "<Metric(metric_id='%d', metric_name='%s', slicer_name='%s', "
            "run_name='%s', sql_constraint='%s', metric_info_label='%s', metric_datafile='%s')>"
        ) % (
            self.metric_id,
            self.metric_name,
            self.slicer_name,
            self.run_name,
            self.sql_constraint,
            self.metric_info_label,
            self.metric_datafile,
        )


class VersionRow(Base):
    """"""

    __tablename__ = "version"
    version_id = Column(Integer, primary_key=True)
    version = Column(String)
    run_date = Column(String)

    def __repr__(self):
        return ("<Version(version='%s', run_date='%s')>") % (
            self.version,
            self.run_date,
        )


class DisplayRow(Base):
    """
    Define contents and format of the displays table.

    (Table to list the display properties for each metric.)
    """

    __tablename__ = "displays"
    display_id = Column(Integer, primary_key=True)
    metric_id = Column(Integer, ForeignKey("metrics.metric_id"))
    # Group for displaying metric (in webpages).
    display_group = Column(String)
    # Subgroup for displaying metric.
    display_subgroup = Column(String)
    # Order to display metric (within subgroup).
    display_order = Column(Float)
    # The figure caption.
    display_caption = Column(String)
    metric = relationship("MetricRow", backref=backref("displays", order_by=display_id))

    def __rep__(self):
        return (
            "<Display(display_group='%s', display_subgroup='%s', "
            "display_order='%.1f', display_caption='%s')>"
            % (
                self.display_group,
                self.display_subgroup,
                self.display_order,
                self.display_caption,
            )
        )


class PlotRow(Base):
    """
    Define contents and format of plot list table.

    (Table to list all plots,
    link them to relevant metrics in MetricList,
    and provide info on filename).
    """

    __tablename__ = "plots"
    # Define columns in plot list table.
    plot_id = Column(Integer, primary_key=True)
    # Matches metricID in MetricList table.
    metric_id = Column(Integer, ForeignKey("metrics.metric_id"))
    plot_type = Column(String)
    plot_file = Column(String)
    metric = relationship("MetricRow", backref=backref("plots", order_by=plot_id))

    def __repr__(self):
        return "<Plot(metric_id='%d', plot_type='%s', plot_file='%s')>" % (
            self.metric_id,
            self.plot_type,
            self.plot_file,
        )


class SummaryStatRow(Base):
    """
    Define contents and format of the summary statistics table.

    (Table to list and link summary stats to relevant metrics in MetricList,
    and provide summary stat name, value and potentially a comment).
    """

    __tablename__ = "summarystats"
    # Define columns in summary table.
    stat_id = Column(Integer, primary_key=True)
    # Matches metricID in MetricList table.
    metric_id = Column(Integer, ForeignKey("metrics.metric_id"))
    summary_name = Column(String)
    summary_value = Column(Float)
    metric = relationship("MetricRow", backref=backref("summarystats", order_by=stat_id))

    def __repr__(self):
        return "<SummaryStat(metric_id='%d', summary_name='%s', summary_value='%f')>" % (
            self.metric_id,
            self.summary_name,
            self.summary_value,
        )


class ResultsDb:
    """ResultsDb is a sqlite database containing information on the metrics
    run via MAF, the plots created, the display information (such as captions),
    and any summary statistics output.
    """

    def __init__(self, out_dir=None, database=None, verbose=False, timeout=180):
        """
        Set up the resultsDb database.
        """
        # We now require results_db to be a sqlite file (for simplicity).
        # Leaving as attribute though.
        self.driver = "sqlite"
        # connecting to non-existent database creates it automatically
        if database is None:
            # Using default value for database name, should specify directory.
            if out_dir is None:
                out_dir = "."
            # Check for output directory, make if needed.
            if not os.path.isdir(out_dir):
                try:
                    os.makedirs(out_dir)
                except OSError as msg:
                    raise OSError(
                        msg,
                        "\n  (If this was the database file (not out_dir), "
                        'remember to use kwarg "database")',
                    )
            self.database = os.path.join(out_dir, "resultsDb_sqlite.db")
        else:
            # Using non-default database, but may also specify directory root.
            if out_dir is not None:
                database = os.path.join(out_dir, database)
            self.database = database
        # If this is a new file, then we should record date and version later.
        needs_version = not os.path.isfile(self.database)

        # Connect to the specified file;
        # this will create the database if it doesn't exist.
        already_file = os.path.isfile(self.database)
        db_address = url.URL.create(self.driver, database=self.database)

        engine = create_engine(db_address, echo=verbose, connect_args={"timeout": timeout})
        self.Session = sessionmaker(bind=engine)
        self.open()
        # Create the tables, if they don't already exist.
        if not already_file:
            try:
                Base.metadata.create_all(engine)
            except DatabaseError:
                raise ValueError(
                    "Cannot create a %s database at %s. Check directory exists."
                    % (self.driver, self.database)
                )
        self.slen = 1024
        # Check if we have a database matching this schema
        # (with metric_info_label)
        query = text("select * from metrics limit 1")
        cols = self.session.execute(query)._metadata.keys
        if "sql_constraint" not in cols:
            self.update_database()

        # record the version and date MAF was run with/on
        if needs_version:
            vers = rsVersion.__version__
            run_date = datetime.datetime.now().strftime("%Y-%m-%d")
            versioninfo = VersionRow(version=vers, run_date=run_date)
            self.session.add(versioninfo)
            self.session.commit()

        self.close()

    def update_database(self):
        """Update the results_db from 'metricMetaData' to 'metric_info_label'
        and now also changing the camel case to snake case
        (metricId to metric_id, etc.).

        This updates results_db to work with the current version of MAF,
        including RunComparison and showMaf.
        There is also a 'downgrade_database' to revert to the older style
        with 'metricMetadata.
        """
        warnings.warn(
            "Updating database to match new schema."
            "Undo with self.downgrade_database if necessary (for older maf versions)."
        )
        query = text("select * from metrics limit 1")
        cols = self.session.execute(query)._metadata.keys
        if "metricMetadata" in cols:
            # if it's very old
            query = text("alter table metrics rename column metricMetadata to metric_info_label")
            self.session.execute(query)
            self.session.commit()
        # Check for metricId vs. metric_id separately,
        # as this change happened independently.
        if "metricId" in cols:
            for table in ["metrics", "summarystats", "plots", "displays"]:
                query = text(f"alter table {table} rename column metricId to metric_id")
                self.session.execute(query)
            self.session.commit()
        # Update to newest schema.
        if "run_name" not in cols:
            updates = {
                "metricName": "metric_name",
                "slicerName": "slicer_name",
                "simDataName": "run_name",
                "sqlConstraint": "sql_constraint",
                "metricInfoLabel": "metric_info_label",
                "metricDataFile": "metric_datafile",
            }
            for old, new in updates.items():
                query = text(f"alter table metrics rename column {old} to {new}")
                self.session.execute(query)
            self.session.commit()
            # update summarystat table
            updates = {
                "statId": "stat_id",
                "summaryName": "summary_name",
                "summaryValue": "summary_value",
            }
            for old, new in updates.items():
                query = text(f"alter table summarystats rename column {old} to {new}")
                self.session.execute(query)
            self.session.commit()
            # update plot table
            updates = {
                "plotId": "plot_id",
                "plotType": "plot_type",
                "plotFile": "plot_file",
            }
            for old, new in updates.items():
                query = text(f"alter table plots rename column {old} to {new}")
                self.session.execute(query)
            self.session.commit()
            # update display table
            updates = {
                "displayId": "display_id",
                "displayGroup": "display_group",
                "displaySubgroup": "display_subgroup",
                "displayOrder": "display_order",
                "displayCaption": "display_caption",
            }
            for old, new in updates.items():
                query = text(f"alter table displays rename column {old} to {new}")
                self.session.execute(query)
            self.session.commit()
            updates = {
                "verId": "version_id",
                "rundate": "run_date",
            }
            for old, new in updates.items():
                query = text(f"alter table version rename column {old} to {new}")
                self.session.execute(query)
            self.session.commit()

    def downgrade_database(self):
        """
        Downgrade resultsDb to work with v0.10<MAF< v1.0
        There is also a 'upgradeDatabase' to update to the newer
        style with 'metric_info_label.
        """
        self.open()
        warnings.warn("Downgrading MAF resultsDb to run with MAF < v1.0")
        updates = {
            "metricId": "metric_id",
            "metricName": "metric_name",
            "slicerName": "slicer_name",
            "simDataName": "run_name",
            "sqlConstraint": "sql_constraint",
            "metricInfoLabel": "metric_info_label",
            "metricDataFile": "metric_datafile",
        }
        for old, new in updates.items():
            query = text(f"alter table metrics rename column {new} to {old}")
            self.session.execute(query)
        self.session.commit()
        # update summarystat table
        updates = {
            "statId": "stat_id",
            "metricId": "metric_id",
            "summaryName": "summary_name",
            "summaryValue": "summary_value",
        }
        for old, new in updates.items():
            query = text(f"alter table summarystats rename column {new} to {old}")
            self.session.execute(query)
        self.session.commit()
        # update plot table
        updates = {
            "plotId": "plot_id",
            "metricId": "metric_id",
            "plotType": "plot_type",
            "plotFile": "plot_file",
        }
        for old, new in updates.items():
            query = text(f"alter table plots rename column {new} to {old}")
            self.session.execute(query)
        self.session.commit()
        # update display table
        updates = {
            "displayId": "display_id",
            "metricId": "metric_id",
            "displayGroup": "display_group",
            "displaySubgroup": "display_subgroup",
            "displayOrder": "display_order",
            "displayCaption": "display_caption",
        }
        for old, new in updates.items():
            query = text(f"alter table displays rename column {new} to {old}")
            self.session.execute(query)
        self.session.commit()
        updates = {
            "verId": "version_id",
            "rundate": "run_date",
        }
        for old, new in updates.items():
            query = text(f"alter table version rename column {new} to {old}")
            self.session.execute(query)
        self.session.commit()
        self.close()

    def open(self):
        """
        Open connection to database
        """
        self.session = self.Session()
        self.session.expire_on_commit = False

    def close(self):
        """
        Close connection to database.
        """
        self.session.close()

    def update_metric(
        self,
        metric_name,
        slicer_name,
        run_name,
        sql_constraint,
        metric_info_label,
        metric_datafile,
    ):
        """
        Add a row to or update a row in the metrics table.

        Parameters
        ----------
        metric_name : `str`
            Name of the Metric
        slicer_name : `str`
            Name of the Slicer
        run_name : `str`
            Name of the simulation (run_name, simName, run_name..)
        sql_constraint : `str`
            sql_constraint relevant for the metric bundle
        metric_info_label : `str`
            Information associated with the metric.
            Could be derived from the sql_constraint or could
            be a more descriptive version, specified by the user.
        metric_datafile : `str`
            The data file the metric bundle output is stored in.

        Returns
        -------
        metric_id : `int`
            The Id number of this metric in the metrics table.

        If same metric (same metric_name, slicer_name, run_name,
        sql_constraint, infoLabel) already exists, it does nothing.
        """
        self.open()
        if run_name is None:
            run_name = "NULL"
        if sql_constraint is None:
            sql_constraint = "NULL"
        if metric_info_label is None:
            metric_info_label = "NULL"
        if metric_datafile is None:
            metric_datafile = "NULL"
        # Check if metric has already been added to database.
        prev = (
            self.session.query(MetricRow)
            .filter_by(
                metric_name=metric_name,
                slicer_name=slicer_name,
                run_name=run_name,
                metric_info_label=metric_info_label,
                sql_constraint=sql_constraint,
            )
            .all()
        )
        if len(prev) == 0:
            metricinfo = MetricRow(
                metric_name=metric_name,
                slicer_name=slicer_name,
                run_name=run_name,
                sql_constraint=sql_constraint,
                metric_info_label=metric_info_label,
                metric_datafile=metric_datafile,
            )
            self.session.add(metricinfo)
            self.session.commit()
        else:
            metricinfo = prev[0]
        self.close()

        return metricinfo.metric_id

    def update_display(self, metric_id, display_dict, overwrite=True):
        """
        Add a row to or update a row in the displays table.

        Parameters
        ----------
        metric_id : `int`
            The metricID for this metric bundle in the metrics table
        display_dict : `dict`
            Dictionary containing the display info
            (group/subgroup/order/caption)
        overwrite : `bool`, opt
            Replaces existing row with same metric_id if overwrite is True.
        """
        # Because we want to maintain 1-1 relationship between
        # metric_id's and display_dict's: First check if a display line
        # is present with this metricID.
        self.open()
        displayinfo = self.session.query(DisplayRow).filter_by(metric_id=metric_id).all()
        if len(displayinfo) > 0:
            if overwrite:
                for d in displayinfo:
                    self.session.delete(d)
            else:
                return
        # Then go ahead and add new display_dict.
        for k in display_dict:
            if display_dict[k] is None:
                display_dict[k] = "NULL"
        keys = ["group", "subgroup", "order", "caption"]
        for k in keys:
            if k not in display_dict:
                display_dict[k] = "NULL"
        if display_dict["order"] == "NULL":
            display_dict["order"] = 0
        display_group = display_dict["group"]
        display_subgroup = display_dict["subgroup"]
        display_order = display_dict["order"]
        display_caption = display_dict["caption"]
        if display_caption.endswith("(auto)"):
            display_caption = display_caption.replace("(auto)", "", 1)
        displayinfo = DisplayRow(
            metric_id=metric_id,
            display_group=display_group,
            display_subgroup=display_subgroup,
            display_order=display_order,
            display_caption=display_caption,
        )
        self.session.add(displayinfo)
        self.session.commit()
        self.close()

    def update_plot(self, metric_id, plot_type, plot_file, overwrite=False):
        """
        Add a row to or update a row in the plot table.

        Parameters
        ----------
        metric_id : `int`
            The metric Id of this metric bundle in the metrics table
        plot_type : `str`
            The type of this plot (oneDbinned data, healpix map, etc.)
        plot_file : `str`
            The filename for this plot
        overwrite : `bool`
            If True, replaces existing row.
            If False, an additional plot is added to the output (e.g. with
            a different range of color values, etc).
        """
        self.open()
        plotinfo = self.session.query(PlotRow).filter_by(metric_id=metric_id, plot_type=plot_type).all()
        if len(plotinfo) > 0 and overwrite:
            for p in plotinfo:
                self.session.delete(p)
        plotinfo = PlotRow(metric_id=metric_id, plot_type=plot_type, plot_file=plot_file)
        self.session.add(plotinfo)
        self.session.commit()
        self.close()

    def update_summary_stat(self, metric_id, summary_name, summary_value, ntry=3, pause_time=100):
        """
        Add a row to or update a row in the summary statistic table.

        Most summary statistics will be a simple name (string) + value (float)
        pair. For special summary statistics which must return multiple values,
        the base name can be provided as 'name', together with a np.ndarray as
        'value', where the array also has 'name' and 'value' columns
        (and each name/value pair is then saved as a summary statistic
        associated with this same metric_id).

        Parameters
        ----------
        metric_id : `int`
            The metric Id of this metric bundle
        summary_name : `str`
            The name of this summary statistic
        summary_value: : `float` or `np.ndarray`
            The value for this summary statistic.
            If this is a np.ndarray, then it should also have 'name' and
            'value' columns to save each value to rows in the summary stats.
        ntry : `int`, opt
            The number of times to retry if database is locked.
        pause_time : `int`, opt
            Time to wait until trying again.
        """
        # Allow for special summary statistics which return data in a
        # np.ndarray with 'name' and 'value' columns.
        self.open()
        tries = 0
        if isinstance(summary_value, np.ndarray) and summary_value.dtype.names is not None:
            if ("name" in summary_value.dtype.names) and ("value" in summary_value.dtype.names):
                for value in summary_value:
                    sSuffix = value["name"]
                    if isinstance(sSuffix, bytes):
                        sSuffix = sSuffix.decode("utf-8")
                    else:
                        sSuffix = str(sSuffix)
                    summarystat = SummaryStatRow(
                        metric_id=metric_id,
                        summary_name=summary_name + " " + sSuffix,
                        summary_value=value["value"],
                    )
                    success = False
                    # This can hit a locked database if running in parallel
                    # have it try a few times before actually failing
                    # since nothing should be writing for a long time.
                    while (not success) & (tries < ntry):
                        try:
                            self.session.add(summarystat)
                            self.session.commit()
                            success = True
                        except OperationalError:
                            tries += 1
                            time.sleep(pause_time)
            else:
                warnings.warn("Warning! Cannot save non-conforming summary statistic.")
        # Most summary statistics will be simple floats.
        else:
            if isinstance(summary_value, float) or isinstance(summary_value, int):
                summarystat = SummaryStatRow(
                    metric_id=metric_id,
                    summary_name=summary_name,
                    summary_value=summary_value,
                )
                self.session.add(summarystat)
                self.session.commit()
            else:
                warnings.warn("Warning! Cannot save summary statistic that is not a simple float or int")
        self.close()

    def get_metric_id(self, metric_name, slicer_name=None, metric_info_label=None, run_name=None):
        """Find metric bundle Ids from the metric table.

        Parameters
        ----------
        metric_name : `str`
            Name of the Metric
        slicer_name : `str`, opt
            Name of the Slicer to match
        metric_info_label : `str`, opt
            Metadata value to match
        run_name : `str`, opt
            Name of the simulation (run_name) to match

        Returns
        -------
        metric_id : `list` of `int`
            List of matching metric_ids
        """
        self.open()
        metric_id = []
        query = self.session.query(
            MetricRow.metric_id,
            MetricRow.metric_name,
            MetricRow.slicer_name,
            MetricRow.metric_info_label,
            MetricRow.run_name,
        ).filter(MetricRow.metric_name == metric_name)
        if slicer_name is not None:
            query = query.filter(MetricRow.slicer_name == slicer_name)
        if metric_info_label is not None:
            query = query.filter(MetricRow.metric_info_label == metric_info_label)
        if run_name is not None:
            query = query.filter(MetricRow.run_name == run_name)
        query = query.order_by(MetricRow.slicer_name, MetricRow.metric_info_label)
        for m in query:
            metric_id.append(m.metric_id)
        self.close()
        return metric_id

    def get_metric_id_like(
        self,
        metric_name_like=None,
        slicer_name_like=None,
        metric_info_label_like=None,
        run_name=None,
    ):
        """Find metric bundle Ids from the metric table,
        but search for names 'like' the values.
        (instead of a strict match from get_metric_id).

        Parameters
        ----------
        metric_name : `str`
            Partial name of the Metric
        slicer_name : `str`, opt
            Partial name of the Slicer to match
        metric_info_label : `str`, opt
            Partial info_label value to match
        run_name : `str`, opt
            Name of the simulation (run_name) to match (exact)

        Returns
        -------
        metric_id : `list` of `int`
            List of matching metric_ids
        """
        self.open()
        metric_id = []
        query = self.session.query(
            MetricRow.metric_id,
            MetricRow.metric_name,
            MetricRow.slicer_name,
            MetricRow.metric_info_label,
            MetricRow.run_name,
        )
        if metric_name_like is not None:
            query = query.filter(MetricRow.metric_name.like(f"%{str(metric_name_like)}%"))
        if slicer_name_like is not None:
            query = query.filter(MetricRow.slicer_name.like(f"%{str(slicer_name_like)}%"))
        if metric_info_label_like is not None:
            query = query.filter(MetricRow.metric_info_label.like(f"%{str(metric_info_label_like)}%"))
        if run_name is not None:
            query = query.filter(MetricRow.run_name == run_name)
        for m in query:
            metric_id.append(m.metric_id)
        self.close()
        return metric_id

    def get_all_metric_ids(self):
        """
        Return a list of all metric_ids.
        """
        self.open()
        metric_ids = []
        for m in self.session.query(MetricRow.metric_id).all():
            metric_ids.append(m.metric_id)
        self.close()
        return metric_ids

    @staticmethod
    def build_summary_name(metric_name, metric_info_label, slicer_name, summary_stat_name=None):
        """Standardize a complete summary metric name,
        combining the metric + slicer + summary + info_label."""
        if metric_info_label is None:
            metric_info_label = ""
        if slicer_name is None:
            slicer_name = ""
        sName = summary_stat_name
        if sName == "Identity" or sName == "Id" or sName == "Count" or sName is None:
            sName = ""
        slName = slicer_name
        if slName == "UniSlicer":
            slName = ""
        name = " ".join([sName, metric_name, metric_info_label, slName]).rstrip(" ").lstrip(" ")
        name.replace(",", "")
        return name

    def get_summary_stats(
        self,
        metric_id=None,
        summary_name=None,
        summary_name_like=None,
        summary_name_notlike=None,
        with_sim_name=False,
    ):
        """
        Get the summary stats (optionally for metric_id list).
        Optionally, also specify the summary metric name.
        Returns a numpy array of the metric information +
        summary statistic information.

        Parameters
        ----------
        metric_id : `int` or `list` of `int`
            Metric bundle Ids to match from the metric table
        summary_name : `str`, opt
            Match this summary statistic name exactly.
        summary_name_like : `str`, opt
            Partial match to this summary statistic name.
        summary_name_notlike : `str`, opt
            Exclude summary statistics with summary names like this.
        with_sim_name : `bool`, opt
            If True, add the run_name to the returned numpy recarray.

        Returns
        -------
        summarystats : `np.recarray`
            Numpy recarray containing the selected summary stat information.
        """
        if metric_id is not None:
            if not hasattr(metric_id, "__iter__"):
                metric_id = [
                    metric_id,
                ]
        summarystats = []
        self.open()
        # Join the metric table and the summarystat table,
        # based on the metricID (the second filter)
        query = self.session.query(MetricRow, SummaryStatRow).filter(
            MetricRow.metric_id == SummaryStatRow.metric_id
        )
        if metric_id is not None:
            query = query.filter(MetricRow.metric_id.in_(metric_id))
        if summary_name is not None:
            query = query.filter(SummaryStatRow.summary_name == str(summary_name))
        if summary_name_like is not None:
            query = query.filter(SummaryStatRow.summary_name.like(f"%{str(summary_name_like)}%"))
        if summary_name_notlike is not None:
            if isinstance(summary_name_notlike, list):
                for s in summary_name_notlike:
                    query = query.filter(~SummaryStatRow.summary_name.like(f"%{str(s)}%"))
            else:
                query = query.filter(~SummaryStatRow.summary_name.like(f"%{str(summary_name_notlike)}%"))
        for m, s in query:
            long_name = self.build_summary_name(
                m.metric_name, m.metric_info_label, m.slicer_name, s.summary_name
            )
            vals = (
                m.metric_id,
                long_name,
                m.metric_name,
                m.slicer_name,
                m.metric_info_label,
                s.summary_name,
                s.summary_value,
            )
            if with_sim_name:
                vals += (m.run_name,)
            summarystats.append(vals)
        # Convert to numpy array.
        dtype_list = [
            ("metric_id", int),
            ("summary_name", str, self.slen),
            ("metric_name", str, self.slen),
            ("slicer_name", str, self.slen),
            ("metric_info_label", str, self.slen),
            ("summary_metric", str, self.slen),
            ("summary_value", float),
        ]
        if with_sim_name:
            dtype_list += [("run_name", str, self.slen)]
        dtype = np.dtype(dtype_list)
        summarystats = np.array(summarystats, dtype)
        self.close()
        return summarystats

    def get_plot_files(self, metric_id=None, with_sim_name=False):
        """
        Find the metric_id, name, info_label, and all plot info
        (optionally for metric_id list).

        Parameters
        ----------
        metric_id : `int`  `list`, or `None`
            If None, plots for all metrics are returned. Otherwise, only plots
            corresponding to the supplied metric ID or IDs are returned
        with_sim_name : `bool`
            If True, include the run name in the fields returned

        Returns
        -------
        plotFiles : `numpy.recarray`
            ``metric_id``
                The metric ID
            ``metric_name``
                The metric name
            ``metric_info_label``
                info_label extracted from the sql constraint
                (usually the filter)
            ``plot_type``
                The plot type
            ``plot_file``
                The full plot file (pdf by default)
            ``thumb_file``
                A plot thumbnail file name (png)
            ``run_name``
                The name of the run plotted (if `with_sim_name` was `True`)
        """
        if metric_id is None:
            metric_id = self.get_all_metric_ids()
        if not hasattr(metric_id, "__iter__"):
            metric_id = [
                metric_id,
            ]
        self.open()
        plotfiles = []
        for mid in metric_id:
            # Join the metric table and the plot table based on the metricID
            # (the second filter does the join)
            query = (
                self.session.query(MetricRow, PlotRow)
                .filter(MetricRow.metric_id == mid)
                .filter(MetricRow.metric_id == PlotRow.metric_id)
            )
            for m, p in query:
                # The plot_file typically ends with .pdf
                # (but the rest of name can have '.' or '_')
                thumb_file = "thumb." + ".".join(p.plot_file.split(".")[:-1]) + ".png"
                plot_file_fields = (
                    m.metric_id,
                    m.metric_name,
                    m.metric_info_label,
                    p.plot_type,
                    p.plot_file,
                    thumb_file,
                )
                if with_sim_name:
                    plot_file_fields += (m.run_name,)
                plotfiles.append(plot_file_fields)

        # Convert to numpy array.
        dtype_list = [
            ("metric_id", int),
            ("metric_name", str, self.slen),
            ("metric_info_label", str, self.slen),
            ("plot_type", str, self.slen),
            ("plot_file", str, self.slen),
            ("thumb_file", str, self.slen),
        ]

        if with_sim_name:
            dtype_list += [("run_name", str, self.slen)]
        dtype = np.dtype(dtype_list)

        plotfiles = np.array(plotfiles, dtype)
        self.close()
        return plotfiles

    def get_metric_data_files(self, metric_id=None):
        """
        Get the metric data filenames for all or a single metric.
        Returns a list.
        """
        if metric_id is None:
            metric_id = self.get_all_metric_ids()
        self.open()
        if not hasattr(metric_id, "__iter__"):
            metric_id = [
                metric_id,
            ]
        datafiles = []
        for mid in metric_id:
            for m in self.session.query(MetricRow).filter(MetricRow.metric_id == mid).all():
                datafiles.append(m.metric_datafile)
        self.close()
        return datafiles

    def get_metric_info(self, metric_id=None, with_sim_name=False):
        """Get the simple metric info, without display information.

        Parameters
        ----------
        metric_id : `int`  `list`, or `None`
            If None, data for all metrics are returned. Otherwise, only data
            corresponding to the supplied metric ID or IDs are returned
        with_sim_name : `bool`
            If True, include the run name in the fields returned

        Returns
        -------
        plotFiles : `numpy.recarray`
            ``metric_id``
                The metric ID
            ``metric_name``
                The metric name
            ``basemetric_names``
                The base metric names
            ``slicer_name``
                The name of the slicer used in the bundleGroup
            ``sql_constraint``
                The full sql constraint used in the bundleGroup
            ``metric_info_label``
                Metadata extracted from the `sql_constraint`
                (usually the filter)
            ``metric_datafile``
                The file name of the file with the metric data itself.
            ``run_name``
                The name of the run plotted (if `with_sim_name` was `True`)

        """
        if metric_id is None:
            metric_id = self.get_all_metric_ids()
        if not hasattr(metric_id, "__iter__"):
            metric_id = [
                metric_id,
            ]
        self.open()
        metricInfo = []
        for mId in metric_id:
            # Query for all rows in metrics and displays that match metric_ids
            query = self.session.query(MetricRow).filter(MetricRow.metric_id == mId)
            for m in query:
                base_metric_name = m.metric_name.split("_")[0]
                mInfo = (
                    m.metric_id,
                    m.metric_name,
                    base_metric_name,
                    m.slicer_name,
                    m.sql_constraint,
                    m.metric_info_label,
                    m.metric_datafile,
                )
                if with_sim_name:
                    mInfo += (m.run_name,)

                metricInfo.append(mInfo)
        # Convert to numpy array.
        dtype_list = [
            ("metric_id", int),
            ("metric_name", str, self.slen),
            ("base_metric_names", str, self.slen),
            ("slicer_name", str, self.slen),
            ("sql_constraint", str, self.slen),
            ("metric_info_label", str, self.slen),
            ("metric_datafile", str, self.slen),
        ]
        if with_sim_name:
            dtype_list += [("run_name", str, self.slen)]
        dtype = np.dtype(dtype_list)
        metricInfo = np.array(metricInfo, dtype)
        self.close()
        return metricInfo

    def get_metric_display_info(self, metric_id=None):
        """
        Get the contents of the metrics and displays table,
        together with the 'basemetric_name' (optionally, for metric_id list).
        Returns a numpy array of the metric information + display information.

        An underlying assumption here is that all metrics have some
        display info. This may not always be the case.
        """
        if metric_id is None:
            metric_id = self.get_all_metric_ids()
        if not hasattr(metric_id, "__iter__"):
            metric_id = [
                metric_id,
            ]
        self.open()
        metricInfo = []
        for mId in metric_id:
            # Query for all rows in metrics and displays that match metric_ids.
            query = (
                self.session.query(MetricRow, DisplayRow)
                .filter(MetricRow.metric_id == mId)
                .filter(MetricRow.metric_id == DisplayRow.metric_id)
            )
            for m, d in query:
                base_metric_name = m.metric_name.split("_")[0]
                mInfo = (
                    m.metric_id,
                    m.metric_name,
                    base_metric_name,
                    m.slicer_name,
                    m.sql_constraint,
                    m.metric_info_label,
                    m.metric_datafile,
                    d.display_group,
                    d.display_subgroup,
                    d.display_order,
                    d.display_caption,
                )
                metricInfo.append(mInfo)
        # Convert to numpy array.
        dtype = np.dtype(
            [
                ("metric_id", int),
                ("metric_name", np.str_, self.slen),
                ("base_metric_name", np.str_, self.slen),
                ("slicer_name", np.str_, self.slen),
                ("sql_constraint", np.str_, self.slen),
                ("metric_info_label", np.str_, self.slen),
                ("metric_datafile", np.str_, self.slen),
                ("display_group", np.str_, self.slen),
                ("display_subgroup", np.str_, self.slen),
                ("display_order", float),
                ("display_caption", np.str_, self.slen * 10),
            ]
        )
        metricInfo = np.array(metricInfo, dtype)
        self.close()
        return metricInfo

    def get_run_name(self):
        """Return a list of the run_names for the metric bundles in
        the database.
        """
        self.open()
        query = self.session.query(MetricRow.run_name.distinct()).all()
        run_name = []
        for s in query:
            run_name.append(s[0])
        self.close()
        return run_name



__all__ = ("ColInfo",)


from .base_stacker import BaseStacker


class ColInfo:
    """Holds the units and source (stacker or database) locations
    for data columns.

    The stacker classes which will generate stacker columns are tracked here,
    as well as some default units for common opsim columns.

    See ColInfo.unit_dict for the unit information of database columns.
    """

    def __init__(self):
        self.default_data_source = None
        self.default_unit = ""
        self.unit_dict = {
            "filter": "filter",
            "seqnNum": "#",
            "expMJD": "MJD",
            "observationStartMJD": "MJD",
            "observationStartLST": "deg",
            "visitExposureTime": "s",
            "slewTime": "s",
            "slewDist": "rad",
            "rotSkyPos": "deg",
            "rotTelPos": "deg",
            "rawSeeing": "arcsec",
            "finSeeing": "arcsec",
            "FWHMeff": "arcsec",
            "FWHMgeom": "arcsec",
            "seeingFwhmEff": "arcsec",
            "seeingFwhmGeom": "arcsec",
            "seeingFwhm500": "arcsec",
            "seeing": "arcsec",
            "airmass": "X",
            "night": "days",
            "moonRA": "rad",
            "moonDec": "rad",
            "moonAlt": "rad",
            "dist2Moon": "rad",
            "filtSkyBrightness": "mag/sq arcsec",
            "skyBrightness": "mag/sq arcsec",
            "fiveSigmaDepth": "mag",
            "solarElong": "degrees",
        }
        # Go through the available stackers and add any units,
        # and identify their source methods.
        self.source_dict = {}
        for stacker_class in BaseStacker.registry.values():
            stacker = stacker_class()
            for col, units in zip(stacker.cols_added, stacker.units):
                self.source_dict[col] = stacker_class
                self.unit_dict[col] = units
        # Note that a 'unique' list of methods should be built from the
        # resulting returned methods, at whatever point the derived data
        # columns will be calculated.

    def get_units(self, col_name):
        """Return the appropriate units for col_name.

        If no units have been defined for a given column,
        return the default units ('').

        Parameters
        ----------
        col_name : `str`
            The name of the column

        Returns
        -------
        units : `str`
        """
        if col_name is None or col_name not in self.unit_dict:
            return self.default_unit
        else:
            return self.unit_dict[col_name]

    def get_data_source(self, col_name):
        """Identify the appropriate source for a given column.

        For values which are calculated via a stacker, the returned
        value is the stacker class.
        For values which do not have a recorded source or are known to
        be coming from the database, the result is
        self.default_data_source (None), which will be assumed to be
        queryable from the database.

        Parameters
        ----------
        col_name : `str`
            The name of the column.

        Returns
        -------
        data_source : `rubin_sim.maf.stacker` or None
        """
        if col_name in self.source_dict:
            return self.source_dict[col_name]
        else:
            return self.default_data_source



__all__ = (
    "PassMetric",
    "Coaddm5Metric",
    "MaxMetric",
    "AbsMaxMetric",
    "MeanMetric",
    "AbsMeanMetric",
    "MedianMetric",
    "AbsMedianMetric",
    "MinMetric",
    "FullRangeMetric",
    "RmsMetric",
    "RelRmsMetric",
    "SumMetric",
    "CountUniqueMetric",
    "CountMetric",
    "CountRatioMetric",
    "CountSubsetMetric",
    "CountBeyondThreshold",
    "RobustRmsMetric",
    "MaxPercentMetric",
    "AbsMaxPercentMetric",
    "BinaryMetric",
    "FracAboveMetric",
    "FracBelowMetric",
    "PercentileMetric",
    "NoutliersNsigmaMetric",
    "UniqueRatioMetric",
    "MeanAngleMetric",
    "RmsAngleMetric",
    "FullRangeAngleMetric",
    "CountExplimMetric",
    "AngularSpreadMetric",
    "RealMeanMetric",
)

import numpy as np

from .base_metric import BaseMetric

# A collection of commonly used simple metrics,
# operating on a single column and returning a float.

twopi = 2.0 * np.pi


class PassMetric(BaseMetric):
    """Pass the entire dataslice array back to the MetricBundle.

    This is most likely useful while prototyping metrics and wanting to
    just 'get the data at a point in the sky', while using a HealpixSlicer
    or a UserPointSlicer.
    """

    def __init__(self, cols=None, **kwargs):
        if cols is None:
            cols = []
        super(PassMetric, self).__init__(col=cols, metric_dtype="object", **kwargs)

    def run(self, data_slice, slice_point=None):
        return data_slice


class Coaddm5Metric(BaseMetric):
    """Calculate the coadded m5 value.

    Parameters
    ----------
    m5_col : `str`, optional
        Name of the m5 column. Default fiveSigmaDepth.
    metric_name : `str`, optional
        Name to associate with the metric output. Default "CoaddM5".
    filter_name : `str`, optional
        Optionally specify a filter to sub-select visits.
        Default None, does no sub-selection or checking.
    filter_col : `str`, optional
        Name of the filter column.
    units : `str`, optional
        Units for the metric. Default "mag".
    """

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        metric_name="CoaddM5",
        filter_name=None,
        filter_col="Filter",
        units="mag",
        **kwargs,
    ):
        self.filter_name = filter_name
        self.filter_col = filter_col
        self.m5_col = m5_col
        super().__init__(col=m5_col, metric_name=metric_name, units=units, **kwargs)

    @staticmethod
    def coadd(single_visit_m5s):
        return 1.25 * np.log10(np.sum(10.0 ** (0.8 * single_visit_m5s)))

    def run(self, data_slice, slice_point=None):
        if len(data_slice) == 0:
            return self.badval
        if self.filter_name is not None:
            matched = np.where(data_slice[self.filter_col] == self.filter_name)
            coadd = self.coadd(data_slice[matched][self.m5_col])
        else:
            coadd = self.coadd(data_slice[self.m5_col])
        return coadd


class MaxMetric(BaseMetric):
    """Calculate the maximum of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.max(data_slice[self.colname])


class AbsMaxMetric(BaseMetric):
    """Calculate the max of the absolute value of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.max(np.abs(data_slice[self.colname]))


class MeanMetric(BaseMetric):
    """Calculate the mean of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.mean(data_slice[self.colname])


class AbsMeanMetric(BaseMetric):
    """Calculate the mean of the absolute value of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.mean(np.abs(data_slice[self.colname]))


class MedianMetric(BaseMetric):
    """Calculate the median of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.median(data_slice[self.colname])


class AbsMedianMetric(BaseMetric):
    """Calculate the median of the absolute value of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.median(np.abs(data_slice[self.colname]))


class MinMetric(BaseMetric):
    """Calculate the minimum of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.min(data_slice[self.colname])


class FullRangeMetric(BaseMetric):
    """Calculate the range of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.max(data_slice[self.colname]) - np.min(data_slice[self.colname])


class RmsMetric(BaseMetric):
    """Calculate the standard deviation of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.std(data_slice[self.colname])


class RelRmsMetric(BaseMetric):
    """Calculate the relative scatter metric (RMS divided by median)."""

    def run(self, data_slice, slice_point=None):
        return np.std(data_slice[self.colname]) / np.median(data_slice[self.colname])


class SumMetric(BaseMetric):
    """Calculate the sum of a simData column slice."""

    def run(self, data_slice, slice_point=None):
        return np.sum(data_slice[self.colname])


class CountUniqueMetric(BaseMetric):
    """Return the number of unique values."""

    def run(self, data_slice, slice_point=None):
        return np.size(np.unique(data_slice[self.colname]))


class UniqueRatioMetric(BaseMetric):
    """Return the number of unique values divided by the
    total number of values."""

    def run(self, data_slice, slice_point=None):
        ntot = float(np.size(data_slice[self.colname]))
        result = np.size(np.unique(data_slice[self.colname])) / ntot
        return result


class CountMetric(BaseMetric):
    """Count the length of a simData column slice."""

    def __init__(self, col=None, units="#", **kwargs):
        super().__init__(col=col, units=units, **kwargs)
        self.metric_dtype = "int"

    def run(self, data_slice, slice_point=None):
        return len(data_slice[self.colname])


class CountExplimMetric(BaseMetric):
    """Count the number of x second visits.
    Useful for rejecting very short exposures
    and counting 60s exposures as 2 visits.

    Parameters
    ----------
    min_exp : `float`, optional
        Minimum exposure time to consider as a "visit".
        Exposures shorter than this will not be counted (count as 0).
    expected_exp : `float`, optional
        Typical exposure time to expect.
        Exposures longer than this will be counted as
        round(visit exposure time / expected_exp). (i.e. 40s = 1, 50s = 2).
    exp_col : `str`, optional
        Column name to use for visit exposure time.

    Returns
    -------
    value : `int`
        The number of visits longer than min_exp and weighted by expected_exp.
    """

    def __init__(self, min_exp=20.0, expected_exp=30.0, exp_col="visitExposureTime", **kwargs):
        self.min_exp = min_exp
        self.expected_exp = expected_exp
        self.exp_col = exp_col
        if "col" in kwargs:
            del kwargs["col"]
        super().__init__(col=[exp_col], **kwargs)
        self.metric_dtype = "int"

    def run(self, data_slice, slice_point=None):
        nv = data_slice[self.exp_col] / self.expected_exp
        nv[np.where(data_slice[self.exp_col] < self.min_exp)[0]] = 0
        nv = np.round(nv)
        return int(np.sum(nv))


class CountRatioMetric(BaseMetric):
    """Count the length of a column slice, then divide by `norm_val`."""

    def __init__(self, col=None, norm_val=1.0, metric_name=None, units="", **kwargs):
        self.norm_val = float(norm_val)
        if metric_name is None:
            metric_name = "CountRatio %s div %.1f" % (col, norm_val)
        super(CountRatioMetric, self).__init__(col=col, metric_name=metric_name, **kwargs)

    def run(self, data_slice, slice_point=None):
        return len(data_slice[self.colname]) / self.norm_val


class CountSubsetMetric(BaseMetric):
    """Count the length of a column slice which matches `subset`."""

    def __init__(self, col=None, subset=None, units="#", **kwargs):
        super(CountSubsetMetric, self).__init__(col=col, units=units, **kwargs)
        self.metric_dtype = "int"
        self.badval = 0
        self.subset = subset

    def run(self, data_slice, slice_point=None):
        count = len(np.where(data_slice[self.colname] == self.subset)[0])
        return count


class CountBeyondThreshold(BaseMetric):
    """Count the number of entries in a data column above or below
    the `threshold`."""

    def __init__(self, col=None, lower_threshold=None, upper_threshold=None, **kwargs):
        super().__init__(col=col, **kwargs)
        self.lower_threshold = lower_threshold
        self.upper_threshold = upper_threshold

    def run(self, data_slice, slice_point=None):
        # Look for data values which match the criteria for the thresholds
        if self.upper_threshold is None and self.lower_threshold is None:
            count = len(data_slice)
        elif self.upper_threshold is None:
            count = len(np.where(data_slice[self.colname] > self.lower_threshold)[0])
        elif self.lower_threshold is None:
            count = len(np.where(data_slice[self.colname] < self.upper_threshold)[0])
        else:
            count = len(
                np.where(
                    (data_slice[self.colname] > self.lower_threshold)
                    and (data_slice[self.colname] < self.upper_threshold)
                )[0]
            )
        return count


class RobustRmsMetric(BaseMetric):
    """Use the inter-quartile range of the data to estimate the RMS.
    Robust, as this calculation does not include outliers in the distribution.
    """

    def run(self, data_slice, slice_point=None):
        iqr = np.percentile(data_slice[self.colname], 75) - np.percentile(data_slice[self.colname], 25)
        rms = iqr / 1.349  # approximation
        return rms


class MaxPercentMetric(BaseMetric):
    """Return the percent of data which matches the maximum value
    of the data.
    """

    def run(self, data_slice, slice_point=None):
        n_max = np.size(np.where(data_slice[self.colname] == np.max(data_slice[self.colname]))[0])
        percent = n_max / float(data_slice[self.colname].size) * 100.0
        return percent


class AbsMaxPercentMetric(BaseMetric):
    """Return the percent of data which matches the absolute value of the
    max value of the data.
    """

    def run(self, data_slice, slice_point=None):
        max_val = np.abs(np.max(data_slice[self.colname]))
        n_max = np.size(np.where(np.abs(data_slice[self.colname]) == max_val)[0])
        percent = n_max / float(data_slice[self.colname].size) * 100.0
        return percent


class BinaryMetric(BaseMetric):
    """Return 1 if there is data, `badval` otherwise."""

    def run(self, data_slice, slice_point=None):
        if data_slice.size > 0:
            return 1
        else:
            return self.badval


class FracAboveMetric(BaseMetric):
    """Find the fraction of data values above a given `cutoff`."""

    def __init__(self, col=None, cutoff=0.5, scale=1, metric_name=None, **kwargs):
        # Col could just get passed in bundle with kwargs,
        # by explicitly pulling it out first, we support use cases where
        # class instantiated without explicit 'col=').
        if metric_name is None:
            metric_name = "FracAbove %.2f in %s" % (cutoff, col)
        super(FracAboveMetric, self).__init__(col, metric_name=metric_name, **kwargs)
        self.cutoff = cutoff
        self.scale = scale

    def run(self, data_slice, slice_point=None):
        good = np.where(data_slice[self.colname] >= self.cutoff)[0]
        frac_above = np.size(good) / float(np.size(data_slice[self.colname]))
        frac_above = frac_above * self.scale
        return frac_above


class FracBelowMetric(BaseMetric):
    """Find the fraction of data values below a given `cutoff`."""

    def __init__(self, col=None, cutoff=0.5, scale=1, metric_name=None, **kwargs):
        if metric_name is None:
            metric_name = "FracBelow %.2f %s" % (cutoff, col)
        super(FracBelowMetric, self).__init__(col, metric_name=metric_name, **kwargs)
        self.cutoff = cutoff
        self.scale = scale

    def run(self, data_slice, slice_point=None):
        good = np.where(data_slice[self.colname] <= self.cutoff)[0]
        frac_below = np.size(good) / float(np.size(data_slice[self.colname]))
        frac_below = frac_below * self.scale
        return frac_below


class PercentileMetric(BaseMetric):
    """Find the value of a column at a given `percentile`."""

    def __init__(self, col=None, percentile=90, metric_name=None, **kwargs):
        if metric_name is None:
            metric_name = "%.0fth%sile %s" % (percentile, "%", col)
        super(PercentileMetric, self).__init__(col=col, metric_name=metric_name, **kwargs)
        self.percentile = percentile

    def run(self, data_slice, slice_point=None):
        pval = np.percentile(data_slice[self.colname], self.percentile)
        return pval


class NoutliersNsigmaMetric(BaseMetric):
    """Calculate the # of visits less than n_sigma below the mean (n_sigma<0)
    or more than n_sigma above the mean.
    """

    def __init__(self, col=None, n_sigma=3.0, metric_name=None, **kwargs):
        self.n_sigma = n_sigma
        self.col = col
        if metric_name is None:
            metric_name = "Noutliers %.1f %s" % (self.n_sigma, self.col)
        super(NoutliersNsigmaMetric, self).__init__(col=col, metric_name=metric_name, **kwargs)
        self.metric_dtype = "int"

    def run(self, data_slice, slice_point=None):
        med = np.mean(data_slice[self.colname])
        std = np.std(data_slice[self.colname])
        boundary = med + self.n_sigma * std
        # If nsigma is positive, look for outliers above median.
        if self.n_sigma >= 0:
            outsiders = np.where(data_slice[self.colname] > boundary)
        # Else look for outliers below median.
        else:
            outsiders = np.where(data_slice[self.colname] < boundary)
        return len(data_slice[self.colname][outsiders])


def _rotate_angles(angles):
    """Private utility for the '*Angle' Metrics below.

    This takes a series of angles between 0-2pi and rotates them so that the
    first angle is at 0, ensuring the biggest 'gap' is at the end of the
    series.
    This simplifies calculations like the 'mean' and 'rms' or 'fullrange',
    removing the discontinuity at 0/2pi.
    """
    angleidx = np.argsort(angles)
    diffangles = np.diff(angles[angleidx])
    start_to_end = np.array([twopi - angles[angleidx][-1] + angles[angleidx][0]], float)
    if start_to_end < -2.0 * np.pi:
        raise ValueError("Angular metrics expect radians, this seems to be in degrees")
    diffangles = np.concatenate([diffangles, start_to_end])
    maxdiff = np.where(diffangles == diffangles.max())[0]
    if len(maxdiff) > 1:
        maxdiff = maxdiff[-1:]
    if maxdiff == (len(angles) - 1):
        rotation = angles[angleidx][0]
    else:
        rotation = angles[angleidx][maxdiff + 1][0]
    return (rotation, (angles - rotation) % twopi)


class MeanAngleMetric(BaseMetric):
    """Calculate the mean of an angular (degree) column slice.

    'MeanAngle' differs from 'Mean' in that it accounts for wraparound at 2pi.
    """

    def run(self, data_slice, slice_point=None):
        """Calculate mean angle via unit vectors.
        If unit vector 'strength' is less than 0.1,
        then just set mean to 180 degrees
        (as this indicates nearly uniformly distributed angles).
        """
        x = np.cos(np.radians(data_slice[self.colname]))
        y = np.sin(np.radians(data_slice[self.colname]))
        meanx = np.mean(x)
        meany = np.mean(y)
        angle = np.arctan2(meany, meanx)
        radius = np.sqrt(meanx**2 + meany**2)
        mean = angle % twopi
        if radius < 0.1:
            mean = np.pi
        return np.degrees(mean)


class RmsAngleMetric(BaseMetric):
    """Calculate the standard deviation of an angular (degrees) column slice.

    'RmsAngle' differs from 'Rms' in that it accounts for wraparound at 2pi.
    """

    def run(self, data_slice, slice_point=None):
        rotation, angles = _rotate_angles(np.radians(data_slice[self.colname]))
        return np.std(np.degrees(angles))


class FullRangeAngleMetric(BaseMetric):
    """Calculate the full range of an angular (degrees) column slice.

    'FullRangeAngle' differs from 'FullRange' in that it accounts for
    wraparound at 2pi.
    """

    def run(self, data_slice, slice_point=None):
        rotation, angles = _rotate_angles(np.radians(data_slice[self.colname]))
        return np.degrees(angles.max() - angles.min())


class AngularSpreadMetric(BaseMetric):
    """Compute the angular spread statistic which measures
    uniformity of a distribution angles accounting for 2pi periodicity.

    The strategy is to first map angles into unit vectors on the unit circle,
    and then compute the 2D centroid of those vectors.
    A uniform distribution of angles will lead to a distribution of
    unit vectors with mean that approaches the origin.
    In contrast, a delta function distribution of angles leads to a
    delta function distribution of unit vectors with a mean that lands on the
    unit circle.

    The angular spread statistic is then defined as 1 - R,
    where R is the radial offset of the mean
    of the unit vectors derived from the input angles.
    R approaches 1 for a uniform distribution
    of angles and 0 for a delta function distribution of angles.

    The optional parameter `period` may be used to specificy periodicity
    other than 2 pi.
    """

    def __init__(self, col=None, period=2.0 * np.pi, **kwargs):
        # https://en.wikipedia.org/wiki/Directional_statistics
        # #Measures_of_location_and_spread
        # jmeyers314@gmail.com
        self.period = period
        super(AngularSpreadMetric, self).__init__(col=col, **kwargs)

    def run(self, data_slice, slice_point=None):
        # Unit vectors; unwrapped at specified period
        x = np.cos(data_slice[self.colname] * 2.0 * np.pi / self.period)
        y = np.sin(data_slice[self.colname] * 2.0 * np.pi / self.period)
        meanx = np.mean(x)
        meany = np.mean(y)
        # radial offset (i.e., length) of the mean unit vector
        R = np.hypot(meanx, meany)
        return 1.0 - R


class RealMeanMetric(BaseMetric):
    """Calculate the mean of a column with no nans or infs."""

    def run(self, data_slice, slice_point=None):
        return np.mean(data_slice[self.colname][np.isfinite(data_slice[self.colname])])



__all__ = ["TEFF_FIDUCIAL_DEPTH", "TEFF_FIDUCIAL_EXPTIME", "TeffStacker", "compute_teff"]

from collections import defaultdict

import numpy as np

from .base_stacker import BaseStacker

# From reference von Karman 500nm zenith seeing of 0.69"
# median zenith dark seeing from sims_skybrightness_pre
# airmass = 1
# 2 "snaps" of 15 seconds each
# m5_flat_sed sysEngVals from rubin_sim
#   commit 6d03bd49550972e48648503ed60784a4e6775b82 (2021-05-18)
# These include constants from:
#   https://github.com/lsst-pst/syseng_throughputs/blob/master/notebooks/generate_sims_values.ipynb
#   commit 7abb90951fcbc70d9c4d0c805c55a67224f9069f (2021-05-05)
# See https://github.com/lsst-sims/smtn-002/blob/master/notebooks/teff_fiducial.ipynb

# If we do not recognize the filter, return a nan. The np.array(np.nan).item
# is a hack to define a function that retruns a nan without defining a useless
# new function or violating the style guide by using a lambda function.
TEFF_FIDUCIAL_DEPTH = defaultdict(
    np.array(np.nan).item,
    {
        "u": 23.71,
        "g": 24.67,
        "r": 24.24,
        "i": 23.82,
        "z": 23.21,
        "y": 22.40,
    }.items(),
)

TEFF_FIDUCIAL_EXPTIME = 30.0


def compute_teff(m5_depth, filter_name, exptime=None, fiducial_depth=None, teff_base=None, normed=False):
    """Compute the effective exposure time for a limiting magnitude.

    Parameters
    ----------
    m5_depth : `float` or `numpy.ndarray`, (N,)
        The five sigma point source limiting magintude.
    filter_name : `str` or `numpy.ndarray`, (N,)
        The name of the filter.
    exptime : `float` or `numpy.ndarray`, (N,)
        The expsore time (seconds), defaults to TEFF_FIDUCIAL_EXPTIME.
    fiducial_depth: `dict` [`str`, `float`]
        A mapping of filter to fiducial depth.
        Defaults to TEFF_FIDUCIAL_DEPTH.
    teff_base : `float`
        The exposure time (in seconds) corresponding to the exposure depth.
    normed : `bool`
        Normalize against the exposure time, such that a value of 1 corresponds
        to the exposure having been taken at fiducial conditions. Defaults
        to False.

    Returns
    -------
    t_eff : `float`
        Effective expsore time, in seconds (if normed is False) or unitless
        (if normed is true).
    """
    if fiducial_depth is None:
        fiducial_depth = TEFF_FIDUCIAL_DEPTH

    if teff_base is None:
        teff_base = TEFF_FIDUCIAL_EXPTIME

    if exptime is None:
        exptime = TEFF_FIDUCIAL_EXPTIME

    try:
        fiducial_m5_depth = fiducial_depth[filter_name]
    except TypeError:
        # If filter_name is not one value, but an iterable, map it according
        # to fiducial_depth.
        if len(filter_name) != len(m5_depth):
            raise ValueError()

        fiducial_m5_depth = np.array([fiducial_depth[b] for b in filter_name])

    t_eff = teff_base * 10.0 ** (0.8 * (m5_depth - fiducial_m5_depth))

    if normed:
        tau = t_eff / exptime
        return tau
    else:
        return t_eff


class TeffStacker(BaseStacker):
    """Add t_eff column corresponding to fiveSigmaDepth.

    Parameters
    ----------
    m5_col : `str` ('fiveSigmaDepth')
        Colum name for the five sigma limiting depth per pointing.
        Defaults to "fiveSigmaDepth".
    filter_col : `str`
        Defaults to "filter"
    exptime_col : `str`
        The column with the exposure time, defaults to
        "visitExposureTime"
    fiducial_depth: `dict` [`str`, `float`]
        A mapping of filter to fiducial depth.
        Defaults to TEFF_FIDUCIAL_DEPTH.
    fiducial_exptime : `float`
        The exposure time (in seconds) corresponding to the exposure depth.
    normed : `bool`
        Normalize to exporuse time.
    """

    cols_added = ["t_eff"]

    def __init__(
        self,
        m5_col="fiveSigmaDepth",
        filter_col="filter",
        exptime_col="visitExposureTime",
        fiducial_depth=None,
        fiducial_exptime=None,
        normed=False,
        **kwargs,
    ):
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.exptime_col = exptime_col
        self.normed = normed
        self.fiducial_depth = TEFF_FIDUCIAL_DEPTH if fiducial_depth is None else fiducial_depth
        self.fiducial_exptime = TEFF_FIDUCIAL_EXPTIME if fiducial_exptime is None else fiducial_exptime
        self.cols_req = [self.m5_col, self.filter_col, self.exptime_col]
        if self.normed and self.exptime_col not in self.cols_req:
            self.cols_req.append(self.exptime_col)

        self.units = "" if self.normed else "seconds"

    def _run(self, sim_data, cols_present=False):
        if cols_present:
            # Column already present in data; assume it is correct and does not
            # need recalculating.
            return sim_data

        sim_data["t_eff"] = compute_teff(
            sim_data[self.m5_col],
            sim_data[self.filter_col],
            sim_data[self.exptime_col],
            self.fiducial_depth,
            self.fiducial_exptime,
            normed=self.normed,
        )

        return sim_data
